import matplotlib.pyplot as plt
from io import BytesIO
from qtpy.QtGui import QPixmap
from gcoreutils.basic_util import apply
from gcoreutils.convert_util import text_to_sympy, sympy_to_latex


# ==================================================
def _latex2pixmap(latex, size=14, color="black", dpi=300, style="svg", verbose=False):
    if style == "svg":
        size *= 4

    try:
        pixmap = QPixmap()
        fig = plt.figure(figsize=(0.01, 0.01))
        fig.text(0, 0, r"$\displaystyle {}$".format(latex), fontsize=int(0.32 * size), color=color)
        output = BytesIO()
        fig.savefig(output, dpi=dpi, transparent=True, format=style, bbox_inches="tight", pad_inches=0.05)
        plt.close(fig)
        output.seek(0)
        svg = output.read()
        pixmap.loadFromData(svg, style)
        return pixmap
    except RuntimeError:
        if verbose:
            raise RuntimeError(f"latex compile error in {latex}.")
        return None


# ==================================================
def _text2pixmap(text, local=None, check_var=None, rational=False, fill=None, verbose=False):
    ltx = sympy_to_latex(text_to_sympy(text, local, check_var, rational))
    if ltx:
        return latex2pixmap(ltx)
    else:
        return None


# ==================================================
def latex2pixmap(latex, size=14, color="black", dpi=300, style="svg", verbose=False):
    """
    convert latex to pixmap.

    Args:
        latex (str): latex text code.
        size (int, optional): font size.
        color (str, optional): matplotlib color name.
        dpi (int, optional): DPI in creating pixmap.
        style (str, optional): render style, "svg/png/jpg".
        verbose (bool, optional): raise ConversionError, if compile error occurs.

    Returns:
        list or QPixmap: rendered pixmap.

    Notes:
        - if compile error occurs, return None.
        - environment to compile latex is required.
    """
    return apply(lambda ltx: _latex2pixmap(ltx, size, color, dpi, style, verbose), latex)


# ==================================================
def text2pixmap(text, local=None, check_var=None, rational=False, fill=None, verbose=False):
    """
    convert text to pixmap via sympy.

    Args:
        text (str): text to convert.
        local (dict, optional): dict of local variables to use when parsing.
        check_var (list, optional): acceptable variable strings.
        rational (bool, optional): convert float into rational ?
        fill (str): text to fill the empty element.
        verbose (bool, optional): raise ConversionError, if error occurs.

    Returns:
        list or QPixmap: rendered pixmap.

    Notes:
        - if conversion error occurs, return None.
    """
    return apply(lambda t: _text2pixmap(t, local, check_var, rational, fill, verbose), text)



"""
execute QtDraw.
"""
import click
from qtdraw.qt_draw import QtDraw
import os


# ================================================== execute QtDraw
@click.command()
@click.argument("filename", nargs=-1)
def cmd(filename):
    """
    execute QtDraw.

        filename : `.qtdw` file without extension.
    """
    n = len(filename)
    if n < 1:
        filename = None
    elif n > 1:
        exit()
    filename = os.path.abspath(filename[0])

    QtDraw(filename=filename).show()



import pandas as pd
from qtpy.QtCore import QAbstractItemModel, QModelIndex, Qt, QVariant, Signal
from qtdraw.core.tree_item import TreeItem


# ==================================================
class TreeItemModel(QAbstractItemModel):
    """
    tree item model.

    References:
        - https://qiita.com/kenasman/items/794d2874d56d0dc37aea
        - https://ymt-lab.com/post/2020/pyqt5-qtreeview-sample/
        - https://doc.qt.io/qtforpython/overviews/qtwidgets-itemviews-editabletreemodel-example.html
        - https://github.com/baoboa/pyqt5/blob/master/examples/itemviews/editabletreemodel/editabletreemodel.py
    """

    rawDataChanged = Signal(str, int)  # (name, row)

    # ==================================================
    def __init__(self, ds, name, def_value=None, parent=None):
        """
        initialize the class.

        Args:
            ds (DataSet): data set.
            name (str): name of data.
            def_value (list, optional): default value.
            parent (QWidget, optional): parent QWidget.
        """
        super().__init__(parent)

        self.name = name  # name of data
        self.ds = ds  # reference of dataset
        self.header = ds[name].columns.values.tolist()  # header
        self.n_column = len(self.header)  # no of columns

        if def_value is None:
            def_value = ["" for _ in range(self.n_column)]
        self.def_value = def_value

        self.root, self.view, self.group = self.setViewData(ds[name])
        self.apply(self.setDisplayData)

        # self._debug_print()

    # ==================================================
    def setViewData(self, df):
        """
        set view data.

        Args:
            df (DataFrame): data frame.

        Returns:
            tuple: view data.
                - TreeItem: root of tree data.
                - DataFrame: data frame for view role.
                - list: data for group top.
        """
        view = df.copy()

        group = df.groupby(df.columns[0])
        child = [i.tolist() for i in group.groups.values()]
        root = TreeItem.create(child)
        gview = [[None for _ in range(len(df.columns))] for _ in range(len(group.groups))]

        return root, view, gview

    # ==================================================
    def apply(self, f, index=QModelIndex()):
        """
        apply function to all indices.

        Args:
            f (function): a function to apply, f(index).
            index (QModelIndex, optional): start index.
        """
        if index.isValid():
            for column in range(self.columnCount(index)):
                f(self.index(index.row(), column, index.parent()))

        if not self.hasChildren(index) or index.flags() & Qt.ItemNeverHasChildren:
            return

        rows = self.rowCount(index)
        for i in range(rows):
            self.apply(f, self.index(i, 0, index))

    # ==================================================
    def rowCount(self, parent=QModelIndex()):
        """
        number of rows for given parent index.

        Args:
            parent (QModelIndex, optional): parent index.

        Returns:
            int: number of rows.
        """
        if not parent.isValid():
            pitem = self.root
        else:
            pitem = parent.internalPointer()
        return pitem.n_child()

    # ==================================================
    def columnCount(self, parent=QModelIndex()):
        """
        number of columns for given parent index.

        Args:
            parent (QModelIndex, optional): parent index.

        Returns:
            int: number of columns.
        """
        return self.n_column

    # ==================================================
    def flags(self, index):
        """
        flag setting.

        Args:
            index (QModelIndex): index

        Returns:
            Qt.Flags: add Editable.
        """
        if not index.isValid():
            return 0

        return Qt.ItemIsEditable | super().flags(index)

    # ==================================================
    def parent(self, index):
        """
        parent index for given index.

        Args:
            index (QModelIndex): index.

        Returns:
            QModelIndex: parent index. For invalid index, return QModelIndex().
        """
        if not index.isValid():
            return QModelIndex()
        citem = index.internalPointer()
        pitem = citem.parent()
        if pitem == self.root:
            return QModelIndex()
        return self.createIndex(pitem.row(), 0, pitem)

    # ==================================================
    def index(self, row, column, parent=QModelIndex()):
        """
        index obtained from row, column, and parent.

        Args:
            row (int): row in parent.
            column (int): column in parent.
            parent (QModelIndex, optional): parent index.

        Returns:
            QModelIndex: index. For invalid index, return QModelIndex().
        """
        if not parent.isValid():
            pitem = self.root
        else:
            pitem = parent.internalPointer()
        citem = pitem.child(row)
        if citem:
            return self.createIndex(row, column, citem)
        else:
            return QModelIndex()

    # ==================================================
    def absIndex(self, index):
        """
        absolute index of data frame from index.

        Args:
            index (QModelIndex): index.

        Returns:
            tuple: absolute index (table-row, table-column, level). For invalid index, return None.
        """
        if not index.isValid():
            return None, None, None

        item = index.internalPointer()
        if item.n_child() == 0:
            row = item.data()
        else:
            row = item.child(0).data()
        column = index.column()
        level = item.level()

        return row, column, level

    # ==================================================
    def childIndex(self, index):
        """
        child and parent indices.

        Args:
            index (QModelIndex): index.

        Returns:
            list: (child and) parent indices.
        """
        indices = [self.index(i, index.column(), index) for i in range(self.rowCount(index))]
        indices += [index]

        return indices

    # ==================================================
    def data(self, index, role):
        """
        data for given index and role.

        Args:
            index (QModelIndex): index.
            role (Qt.Role): role, Qt.DisplayRole, Qt.EditRole, Qt.UserRole.

        Returns:
            Any: data for given index and role. For invalid role/index, return QVariant().

        Notes:
            - return display data for DisplayRole and EditRole, otherwise return raw data.
        """
        if role != Qt.DisplayRole and role != Qt.EditRole and role != Qt.UserRole:
            return QVariant()

        row, column, level = self.absIndex(index)
        if row is None or level == 0 or level > 2:
            return QVariant()

        if role == Qt.UserRole:
            return self.ds[self.name].iat[row, column]

        if level == 1:
            return self.group[index.row()][column]

        return self.view.iat[row, column]

    # ==================================================
    def headerData(self, section, orientation=Qt.Horizontal, role=Qt.DisplayRole):
        """
        header data.

        Args:
            section (int): column.
            orientation (Qt.Orientation, optional): only for Horizontal.
            role (Qt.Role, optional): only for DisplayRole.

        Returns:
            Any: header column of data frame. For invalid section, return None.
        """
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.header[section]

        return None

    # ==================================================
    def setData(self, index, value, role=Qt.EditRole, emit=True):
        """
        set data.

        Args:
            index (QModelIndex): index.
            value (Any): value to set.
            role (Qt.Role, optional): only for Edit.

        Returns:
            bool: return True if sucess, otherwise False.
        """
        if role != Qt.EditRole:
            return False

        row, column, level = self.absIndex(index)
        if row is None or level == 0 or level > 2:
            return False

        index = self.childIndex(index)

        single = len(index) == 1
        for index1 in index:
            row, column, level = self.absIndex(index1)
            if row is not None and (level == 2 or single):
                self.ds[self.name].iat[row, column] = value
                if emit:
                    self.rawDataChanged.emit(self.name, row)
            self.setDisplayData(index1)
            if level == 2:
                self.setDisplayData(index1.parent().siblingAtColumn(index1.column()))

        return True

    # ==================================================
    def setDisplayData(self, index):
        """
        set data for display.

        Args:
            index (QModelIndex): index.

        Notes:
            - self.view and self.group for corresponding index are modified for display.
        """
        row, column, level = self.absIndex(index)
        if row is None:
            return

        item = self.setDisplayDataRaw(index)
        if level == 1:
            self.group[index.row()][column] = item
        else:
            self.view.iat[row, column] = item

    # ==================================================
    def setDisplayDataRaw(self, index):
        """
        function to obtain data for display from raw data.

        Args:
            index (QIndexModel): index.

        Returns:
            Any: data for display.
        """
        # return str((index.row(), index.column())) + " => " + str(self.absIndex(index))
        data = index.data(role=Qt.UserRole)
        return str(data)

    # ==================================================
    def _debug_print(self):
        """
        print tree structure for deubug.
        """

        def debug(index):
            row, column, level = self.absIndex(index)
            if row is not None:
                if level == 1:
                    g = "group"
                else:
                    g = "orig"
                print(g, ": (", index.row(), index.column(), ") => (", row, column, ")", index.data())

        self.apply(debug)
        print("=====")

    # ==================================================
    def appendRow(self, index=QModelIndex(), name=None):
        """
        append row.

        Args:
            index (QModelIndex, optional): index.
            key (str, optional): name.

        Returns:
            bool: True if insert is success.

        Notes:
            - inserted row is simply appended at the last raw of data frame.
        """
        index = index.siblingAtColumn(0)

        # add row data.
        add = self.def_value.copy()
        null = [None for _ in range(self.columnCount())]
        if not index.isValid():  # add
            names = [self.index(row, 0).data(Qt.UserRole) for row in range(self.root.n_child())]
            add[0] = name
            if name not in names:
                self.group.append(null)
            else:
                grow = names.index(name)
                index = self.index(grow, 0)
        else:  # insert
            add[0] = self.data(index, Qt.UserRole)
        last_row = len(self.ds[self.name])
        self.ds.append(self.name, add)
        self.view = pd.concat([self.view, pd.DataFrame([null], columns=self.header)], ignore_index=True)

        # set index.
        if not index.isValid():  # root
            pitem = self.root
            row = pitem.n_child()
            self.beginInsertRows(index, row, row)
            item = TreeItem(last_row, pitem)
            pitem.append(item)
            self.endInsertRows()
            index = self.index(row, 0)
        else:
            _, _, level = self.absIndex(index)
            if level == 1:  # for group
                pitem = index.internalPointer()
                row = pitem.n_child()
                if row == 0:  # no child
                    self.beginInsertRows(index, 0, 1)
                    item = TreeItem(pitem.data(), pitem)
                    pitem.append(item)
                    item = TreeItem(last_row, pitem)
                    pitem.append(item)
                    pitem.setData(None)
                    self.endInsertRows()
                    index1 = self.index(0, 0, index)
                    for column in range(self.columnCount()):
                        index1 = index1.siblingAtColumn(column)
                        self.setDisplayData(index1)
                    index = self.index(1, 0, index)
                else:  # with child
                    self.beginInsertRows(index, row, row)
                    item = TreeItem(last_row, pitem)
                    pitem.append(item)
                    self.endInsertRows()
                    index = self.index(row, 0, index)
            else:  # for child
                pitem = index.parent().internalPointer()
                row = pitem.n_child()
                self.beginInsertRows(index.parent(), row, row)
                item = TreeItem(last_row, pitem)
                pitem.append(item)
                self.endInsertRows()
                index = self.index(row, 0, index.parent())

        # set data.
        for column, value in enumerate(add):
            index = index.siblingAtColumn(column)
            self.setData(index, value, emit=False)

        self.rawDataChanged.emit(self.name, self.absIndex(index)[0])

        return True

    # ==================================================
    def removeRow(self, index=QModelIndex()):
        """
        remove row.

        Args:
            parent (QModelIndex), optional): parent index.

        Returns:
            bool: True if remove is success.

        Notes:
            - removed row is set simply as (first column)="" instead of removing row itself in raw data frame.
        """
        index = index.siblingAtColumn(0)
        row, _, level = self.absIndex(index)

        parent = index.parent()
        if not parent.isValid():
            pitem = self.root
        else:
            pitem = parent.internalPointer()

        if level == 1:  # for group
            self.beginRemoveRows(parent, index.row(), index.row())
            del self.group[index.row()]
            item = index.internalPointer()
            if item.n_child() == 0:
                self.ds[self.name].iloc[row, 0] = ""
                self.view.iloc[row, 0] = ""
                self.rawDataChanged.emit(self.name, self.absIndex(index)[0])
            else:
                for row in range(item.n_child()):
                    arow = item.child(row).data()
                    self.ds[self.name].iloc[arow, 0] = ""
                    self.view.iloc[arow, 0] = ""
                    self.rawDataChanged.emit(self.name, self.absIndex(index.child(row, 0))[0])
            pitem.remove(index.row())
            self.endRemoveRows()
        else:  # for child
            self.ds[self.name].iloc[row, 0] = ""
            self.view.iloc[row, 0] = ""
            self.rawDataChanged.emit(self.name, self.absIndex(index)[0])
            if pitem.n_child() == 2:
                self.beginRemoveRows(parent, 0, 1)
                if index.row() == 0:
                    pitem.setData(pitem.child(1).data())
                else:
                    pitem.setData(pitem.child(0).data())
                pitem.remove(0)
                pitem.remove(0)
                self.endRemoveRows()
            else:
                self.beginRemoveRows(parent, index.row(), index.row())
                pitem.remove(index.row())
                self.endRemoveRows()
            for column in range(self.columnCount()):
                parent1 = parent.siblingAtColumn(column)
                self.setDisplayData(parent1)

        self.apply(self.setDisplayData)  # update all display data.

        return True



# ==================================================
class TreeItem(object):
    """
    tree item.
    """

    # ==================================================
    def __init__(self, data=None, parent=None):
        """
        initialize the class.

        Args:
            data (Any, optional): item data for each row.
            parent (TreeItem, optional): parent item.
        """
        self.pitem = parent  # parent item
        self.idata = data  # item data
        self.citem = []  # list of child item

    # ==================================================
    def __str__(self):
        return f"{self.idata} ({self.n_child()})"

    # ==================================================
    def level(self):
        """
        find item level.

        Returns:
            int: level of item (top is 0).
        """
        l = 0
        p = self.parent()
        while p is not None:
            l += 1
            p = p.parent()
        return l

    # ==================================================
    def parent(self):
        """
        parent item.

        Returns:
            TreeItem: parent item.
        """
        return self.pitem

    # ==================================================
    def data(self):
        """
        item data.

        Returns:
            Any: item data.
        """
        return self.idata

    # ==================================================
    def setData(self, data):
        """
        set item data.

        Args:
            data (Any): item data.
        """
        self.idata = data

    # ==================================================
    def n_child(self):
        """
        number of child items.

        Returns:
            int: number of child items.
        """
        return len(self.citem)

    # ==================================================
    def append(self, item):
        """
        append child.

        Args:
            item (TreeItem): item.
        """
        self.citem.append(item)

    # ==================================================
    def remove(self, row):
        """
        remove child.

        Args:
            row (int): row to remove.

        Returns:
            bool: True if remove is success.
        """
        if row < 0 or row >= self.n_child():
            return False
        del self.citem[row]
        return True

    # ==================================================
    def child(self, row):
        """
        child item of given row.

        Args:
            row (int): row of child item from 0.

        Returns:
            TreeItem: child item.
        """
        if row < 0 or row >= len(self.citem):
            return None
        return self.citem[row]

    # ==================================================
    def row(self):
        """
        row of item in parent item.

        Returns:
            int: row of item.
        """
        if self.pitem:
            return self.pitem.citem.index(self)
        return 0

    # ==================================================
    def dump(self, indent=4, cr="\n"):
        """
        dump tree.

        Args:
            indent (int, optional): indent.
            cr (str, optional): string for new line.
        """

        def write(top):
            spc = " " * indent * top.level()
            s = spc + str(top) + cr
            for i in range(top.n_child()):
                s += write(top.child(i))
            return s

        s = write(self).strip(" " + cr)
        return s

    # ==================================================
    def add_tree(self, tree):
        """
        add tree data.

        Args:
            tree (list): tree data.
        """
        level = self.level()
        if level == 0 or (type(tree) == list and len(tree) > 1):
            for row, data in enumerate(tree):
                item = TreeItem(None, self)
                self.append(item)
                self.child(row).add_tree(data)
        else:
            if type(tree) == list:
                self.setData(tree[0])
            else:
                self.setData(tree)

    # ==================================================
    @classmethod
    def create(cls, tree):
        """
        create item tree.

        Args:
            tree (list): tree data.

        Returns:
            TreeItem: root item.
        """
        root = TreeItem()
        root.add_tree(tree)

        return root



from PyQt5.QtWidgets import QDialog, QTableWidget, QGridLayout
from qtdraw.core.editable_widget import QtText, QtMath
from qtdraw.core.setting import rcParams


# ==================================================
class TableWidget(QTableWidget):
    # ==================================================
    def __init__(self, data, header, vheader, role, align, parent=None):
        row = len(data)
        if row == 0:
            col = 0
        else:
            col = len(data[0])
        super().__init__(row, col, parent)

        for i, row in enumerate(data):
            for j, item in enumerate(row):
                w = self.get_widget(str(item), role[j], align[j], parent)
                self.setCellWidget(i, j, w)

        if header is not None:
            self.setHorizontalHeaderLabels(header)
        else:
            self.horizontalHeader().hide()
        if vheader is not None:
            self.setVerticalHeaderLabels(vheader)
        else:
            self.verticalHeader().hide()

        self.resizeColumnsToContents()
        self.resizeRowsToContents()

    # ==================================================
    def get_widget(self, item, role, align, parent):
        if role == "math":
            color = rcParams["detail.latex.color"]
            style = rcParams["detail.latex.format"]
            dpi = rcParams["detail.latex.dpi"]
            w = QtMath(
                item,
                style=style,
                color=color,
                dpi=dpi,
                read_only=True,
                align=align,
                parent=parent,
            )
        else:  # text.
            item = str(item).replace("j", "i").strip(" ()")
            w = QtText(item, read_only=True, align=align, parent=parent)
        return w


# ==================================================
class TableDialog(QDialog):
    # ==================================================
    def __init__(
        self, data, title="TableDialog", header=None, vheader=None, role=None, align=None, width=512, height=600, parent=None
    ):
        super().__init__(parent)

        self.setWindowTitle(title)
        self.resize(width, height)

        if role is None:
            role = [""] * len(data)
        if align is None:
            align = ["left"] * len(data)

        table = TableWidget(data, header, vheader, role, align, parent)

        bg = rcParams["detail.table.bg_color"]
        table.setStyleSheet(f"selection-background-color: {bg}")

        layout = QGridLayout(self)
        layout.addWidget(table, 0, 0, 1, 1)

        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        self.setLayout(layout)



from qtpy.QtWidgets import (
    QDialog,
    QTabWidget,
    QWidget,
    QSpacerItem,
    QSizePolicy,
    QComboBox,
    QCheckBox,
    QSpinBox,
    QDoubleSpinBox,
    QLabel,
    QGridLayout,
    QDialogButtonBox,
)
from qtpy.QtCore import Qt
from qtdraw.core.editable_widget import QtColorSelector


# ==================================================
class DialogPreference(QDialog):
    # ==================================================
    def __init__(self, preference, setting, apply_callback, parent=None):
        super().__init__(parent)
        self.setModal(True)
        self.setWindowTitle("Preferences")

        self.preference = preference
        self.setting = setting
        self.callback = apply_callback
        self.pref = preference.copy()
        self.sett = setting.copy()
        self.default = preference.copy()
        self.default_set = setting.copy()

        self.resize(350, 200)

        # tab contents.
        self.tab = QTabWidget(self)
        self.create_tab_label()
        self.create_tab_vector()
        self.create_tab_axis()
        self.create_tab_cell()
        self.create_tab_light()
        self.create_tab_spotlight()

        # buttons.
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel | QDialogButtonBox.Apply)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.buttonBox.clicked.connect(self.apply)

        # main layout
        self.layout = QGridLayout(self)
        self.layout.addWidget(self.tab, 0, 0, 1, 2)
        self.layout.addWidget(self.buttonBox, 1, 0, 2, 1)

    # ==================================================
    def create_tab_label(self):
        self.tab_label = QWidget()
        self.tab.addTab(self.tab_label, "label")

        self.lbl_size = QLabel("font size", self)
        self.spin_size = QSpinBox(self)
        self.spin_size.setMinimum(4)
        self.spin_size.setMaximum(48)
        self.spin_size.setProperty("value", self.pref["label.size"])
        self.lbl_bold = QLabel("bold", self)
        self.check_bold = QCheckBox(self)
        self.check_bold.setChecked(self.pref["label.bold"])
        self.lbl_italic = QLabel("italic", self)
        self.check_italic = QCheckBox(self)
        self.check_italic.setChecked(self.pref["label.italic"])
        self.lbl_color = QLabel("color", self)
        self.combo_color = QtColorSelector(
            self.pref["label.color"],
            color_type="color",
            callback=lambda v: self.pref.update({"label.color": v}),
            parent=self,
        )
        self.combo_color.layout.setContentsMargins(0, 0, 0, 0)
        self.lbl_family = QLabel("font family", self)
        self.combo_font = QComboBox(self)
        self.combo_font.addItems(["arial", "courier", "times"])
        self.combo_font.setCurrentText(self.pref["label.font"])

        spacerItem = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        spacerItem1 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.gridLayout_4 = QGridLayout(self.tab_label)
        self.gridLayout_4.addWidget(self.lbl_size, 0, 0, 1, 1)
        self.gridLayout_4.addWidget(self.spin_size, 0, 1, 1, 1)
        self.gridLayout_4.addItem(spacerItem, 0, 2, 1, 1)
        self.gridLayout_4.addWidget(self.lbl_bold, 1, 0, 1, 1)
        self.gridLayout_4.addWidget(self.check_bold, 1, 1, 1, 1)
        self.gridLayout_4.addWidget(self.lbl_italic, 2, 0, 1, 1)
        self.gridLayout_4.addWidget(self.check_italic, 2, 1, 1, 1)
        self.gridLayout_4.addWidget(self.lbl_color, 3, 0, 1, 1)
        self.gridLayout_4.addWidget(self.combo_color, 3, 1, 1, 1)
        self.gridLayout_4.addWidget(self.lbl_family, 4, 0, 1, 1)
        self.gridLayout_4.addWidget(self.combo_font, 4, 1, 1, 1)
        self.gridLayout_4.addItem(spacerItem1, 5, 0, 1, 1)

        self.spin_size.valueChanged.connect(lambda v: self.pref.update({"label.size": v}))
        self.check_bold.toggled.connect(lambda flag: self.pref.update({"label.bold": flag}))
        self.check_italic.toggled.connect(lambda flag: self.pref.update({"label.italic": flag}))
        self.combo_font.currentTextChanged.connect(lambda v: self.pref.update({"label.font": v}))

    # ==================================================
    def create_tab_vector(self):
        self.tab_vector = QWidget()
        self.tab.addTab(self.tab_vector, "vector")

        self.lbl_tip_l = QLabel("tip length", self)
        self.spin_tip_l = QDoubleSpinBox(self)
        self.spin_tip_l.setMaximum(1.0)
        self.spin_tip_l.setSingleStep(0.05)
        self.spin_tip_l.setProperty("value", self.pref["vector.tip.length"])
        self.spin_tip_l.setFocusPolicy(Qt.NoFocus)
        self.lbl_tip_r = QLabel("tip radius", self)
        self.spin_tip_r = QDoubleSpinBox(self)
        self.spin_tip_r.setMaximum(0.5)
        self.spin_tip_r.setSingleStep(0.01)
        self.spin_tip_r.setProperty("value", self.pref["vector.tip.radius"])
        self.spin_tip_r.setFocusPolicy(Qt.NoFocus)
        self.lbl_shaft_r = QLabel("shaft radius", self)
        self.spin_shaft_r = QDoubleSpinBox(self)
        self.spin_shaft_r.setMaximum(0.5)
        self.spin_shaft_r.setSingleStep(0.01)
        self.spin_shaft_r.setProperty("value", self.pref["vector.shaft.radius"])
        self.spin_shaft_r.setFocusPolicy(Qt.NoFocus)

        spacerItem2 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        spacerItem3 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.gridLayout_5 = QGridLayout(self.tab_vector)
        self.gridLayout_5.addWidget(self.lbl_tip_l, 0, 0, 1, 1)
        self.gridLayout_5.addWidget(self.spin_tip_l, 0, 1, 1, 1)
        self.gridLayout_5.addItem(spacerItem2, 0, 2, 1, 1)
        self.gridLayout_5.addWidget(self.lbl_tip_r, 1, 0, 1, 1)
        self.gridLayout_5.addWidget(self.spin_tip_r, 1, 1, 1, 1)
        self.gridLayout_5.addWidget(self.lbl_shaft_r, 2, 0, 1, 1)
        self.gridLayout_5.addWidget(self.spin_shaft_r, 2, 1, 1, 1)
        self.gridLayout_5.addItem(spacerItem3, 3, 0, 1, 1)

        self.spin_tip_l.valueChanged.connect(lambda v: self.pref.update({"vector.tip.length": round(v, 4)}))
        self.spin_tip_r.valueChanged.connect(lambda v: self.pref.update({"vector.tip.radius": round(v, 4)}))
        self.spin_shaft_r.valueChanged.connect(lambda v: self.pref.update({"vector.shaft.radius": round(v, 4)}))

    # ==================================================
    def create_tab_axis(self):
        self.tab_axis = QWidget()
        self.tab.addTab(self.tab_axis, "axis")

        self.lbl_atype = QLabel("type", self)
        self.combo_atype = QComboBox(self)
        self.combo_atype.addItems(["xyz", "abc", "abc*"])
        self.combo_atype.setCurrentText(self.pref["axis.type"])
        self.lbl_asize = QLabel("size", self)
        self.spin_asize = QSpinBox(self)
        self.spin_asize.setMinimum(10)
        self.spin_asize.setMaximum(24)
        self.spin_asize.setProperty("value", self.pref["axis.size"])
        self.lbl_abold = QLabel("bold", self)
        self.check_abold = QCheckBox(self)
        self.check_abold.setChecked(self.pref["axis.bold"])
        self.lbl_aitalic = QLabel("italic", self)
        self.check_aitalic = QCheckBox(self)
        self.check_aitalic.setChecked(self.pref["axis.italic"])
        self.lbl_aposition = QLabel("position", self)
        self.spin_aposition = QDoubleSpinBox(self)
        self.spin_aposition.setMinimum(0.8)
        self.spin_aposition.setMaximum(1.2)
        self.spin_aposition.setSingleStep(0.1)
        self.spin_aposition.setProperty("value", self.pref["axis.position"])
        self.lbl_ctype = QLabel("crystal", self)
        self.combo_ctype = QComboBox(self)
        self.combo_ctype.addItems(["", "triclinic", "monoclinic", "orthorhombic", "tetragonal", "trigonal", "hexagonal", "cubic"])
        self.combo_ctype.setCurrentText(self.sett["crystal"])

        spacerItema1 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        spacerItema2 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.gridLayout_a = QGridLayout(self.tab_axis)
        self.gridLayout_a.addWidget(self.lbl_atype, 0, 0, 1, 1)
        self.gridLayout_a.addWidget(self.combo_atype, 0, 1, 1, 1)
        self.gridLayout_a.addItem(spacerItema1, 0, 2, 1, 1)
        self.gridLayout_a.addWidget(self.lbl_ctype, 1, 0, 1, 1)
        self.gridLayout_a.addWidget(self.combo_ctype, 1, 1, 1, 1)
        self.gridLayout_a.addWidget(self.lbl_asize, 2, 0, 1, 1)
        self.gridLayout_a.addWidget(self.spin_asize, 2, 1, 1, 1)
        self.gridLayout_a.addWidget(self.lbl_abold, 3, 0, 1, 1)
        self.gridLayout_a.addWidget(self.check_abold, 3, 1, 1, 1)
        self.gridLayout_a.addWidget(self.lbl_aitalic, 4, 0, 1, 1)
        self.gridLayout_a.addWidget(self.check_aitalic, 4, 1, 1, 1)
        self.gridLayout_a.addWidget(self.lbl_aposition, 5, 0, 1, 1)
        self.gridLayout_a.addWidget(self.spin_aposition, 5, 1, 1, 1)
        self.gridLayout_a.addItem(spacerItema2, 6, 0, 1, 1)

        self.combo_atype.currentTextChanged.connect(lambda v: self.pref.update({"axis.type": v}))
        self.spin_asize.valueChanged.connect(lambda v: self.pref.update({"axis.size": v}))
        self.check_abold.toggled.connect(lambda flag: self.pref.update({"axis.bold": flag}))
        self.check_aitalic.toggled.connect(lambda flag: self.pref.update({"axis.italic": flag}))
        self.spin_aposition.valueChanged.connect(lambda v: self.pref.update({"axis.position": round(v, 4)}))
        self.combo_ctype.currentTextChanged.connect(lambda v: self.sett.update({"crystal": v}))

    # ==================================================
    def create_tab_cell(self):
        self.tab_cell = QWidget()
        self.tab.addTab(self.tab_cell, "cell")

        self.gridLayout_2 = QGridLayout(self.tab_cell)
        self.lbl_width = QLabel("width", self)
        self.spin_width = QDoubleSpinBox(self)
        self.spin_width.setMaximum(5.0)
        self.spin_width.setSingleStep(0.1)
        self.spin_width.setProperty("value", self.pref["cell.width"])
        self.spin_width.setFocusPolicy(Qt.NoFocus)
        self.lbl_color_2 = QLabel("color", self)
        self.combo_color_2 = QtColorSelector(
            self.pref["cell.color"],
            color_type="color",
            callback=lambda v: self.pref.update({"cell.color": v}),
            parent=self,
        )
        self.combo_color_2.layout.setContentsMargins(0, 0, 0, 0)
        self.lbl_opacity = QLabel("opacity", self)
        self.spin_opacity = QDoubleSpinBox(self)
        self.spin_opacity.setMaximum(1.0)
        self.spin_opacity.setSingleStep(0.1)
        self.spin_opacity.setProperty("value", self.pref["cell.opacity"])
        self.spin_opacity.setFocusPolicy(Qt.NoFocus)

        spacerItem4 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        spacerItem5 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.gridLayout_2.addWidget(self.lbl_width, 0, 0, 1, 1)
        self.gridLayout_2.addWidget(self.spin_width, 0, 1, 1, 1)
        self.gridLayout_2.addItem(spacerItem4, 0, 2, 1, 1)
        self.gridLayout_2.addWidget(self.lbl_color_2, 1, 0, 1, 1)
        self.gridLayout_2.addWidget(self.combo_color_2, 1, 1, 1, 1)
        self.gridLayout_2.addWidget(self.lbl_opacity, 2, 0, 1, 1)
        self.gridLayout_2.addWidget(self.spin_opacity, 2, 1, 1, 1)
        self.gridLayout_2.addItem(spacerItem5, 3, 0, 1, 1)

        self.spin_width.valueChanged.connect(lambda v: self.pref.update({"cell.width": round(v, 4)}))
        self.spin_opacity.valueChanged.connect(lambda v: self.pref.update({"cell.opacity": round(v, 4)}))

    # ==================================================
    def create_tab_light(self):
        self.tab_light = QWidget()
        self.tab.addTab(self.tab_light, "light")

        self.gridLayout = QGridLayout(self.tab_light)
        self.lbl_intensity = QLabel("intensity", self)
        self.spin_intensity = QDoubleSpinBox(self)
        self.spin_intensity.setMaximum(1.0)
        self.spin_intensity.setSingleStep(0.1)
        self.spin_intensity.setProperty("value", self.pref["light.intensity"])
        self.lbl_prb = QLabel("physics based rendering", self)
        self.check_pbr = QCheckBox(self)
        self.check_pbr.setChecked(self.pref["light.pbr"])
        self.lbl_metallic = QLabel("metallic", self)
        self.spin_metallic = QDoubleSpinBox(self)
        self.spin_metallic.setMaximum(1.0)
        self.spin_metallic.setSingleStep(0.1)
        self.spin_metallic.setProperty("value", self.pref["light.metallic"])
        self.lbl_roughness = QLabel("roughness", self)
        self.spin_roughness = QDoubleSpinBox(self)
        self.spin_roughness.setMaximum(1.0)
        self.spin_roughness.setSingleStep(0.1)
        self.spin_roughness.setProperty("value", self.pref["light.roughness"])
        self.lbl_eye_light = QLabel("eye dome lighting", self)
        self.check_eye_dome_light = QCheckBox(self)
        self.check_eye_dome_light.setChecked(self.pref["light.eye_dome_lighting"])

        spacerItem6 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        spacerItem7 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.gridLayout.addWidget(self.lbl_intensity, 0, 0, 1, 1)
        self.gridLayout.addWidget(self.spin_intensity, 0, 1, 1, 1)
        self.gridLayout.addItem(spacerItem6, 0, 2, 1, 1)
        self.gridLayout.addWidget(self.lbl_prb, 1, 0, 1, 1)
        self.gridLayout.addWidget(self.check_pbr, 1, 1, 1, 1)
        self.gridLayout.addWidget(self.lbl_metallic, 1, 2, 1, 1)
        self.gridLayout.addWidget(self.spin_metallic, 1, 3, 1, 1)
        self.gridLayout.addWidget(self.lbl_roughness, 2, 2, 1, 1)
        self.gridLayout.addWidget(self.spin_roughness, 2, 3, 1, 1)
        self.gridLayout.addWidget(self.lbl_eye_light, 3, 0, 1, 1)
        self.gridLayout.addWidget(self.check_eye_dome_light, 3, 1, 1, 1)
        self.gridLayout.addItem(spacerItem7, 4, 0, 1, 1)

        self.spin_intensity.valueChanged.connect(lambda v: self.pref.update({"light.intensity": round(v, 4)}))
        self.check_pbr.toggled.connect(lambda flag: self.pref.update({"light.pbr": flag}))
        self.spin_metallic.valueChanged.connect(lambda v: self.pref.update({"light.metallic": round(v, 4)}))
        self.spin_roughness.valueChanged.connect(lambda v: self.pref.update({"light.roughness": round(v, 4)}))
        self.check_eye_dome_light.toggled.connect(lambda flag: self.pref.update({"light.eye_dome_lighting": flag}))

    # ==================================================
    def create_tab_spotlight(self):
        self.tab_spotlight = QWidget()
        self.tab.addTab(self.tab_spotlight, "spotlight")

        self.gridLayout_3 = QGridLayout(self.tab_spotlight)
        self.lbl_radius = QLabel("radius", self)
        self.spin_radius = QDoubleSpinBox(self)
        self.spin_radius.setMaximum(5.0)
        self.spin_radius.setSingleStep(0.05)
        self.spin_radius.setProperty("value", self.pref["spotlight.size"])
        self.lbl_color_3 = QLabel("color", self)
        self.combo_color_3 = QtColorSelector(
            self.pref["spotlight.color"],
            color_type="color",
            callback=lambda v: self.pref.update({"spotlight.color": v}),
            parent=self,
        )
        self.combo_color_3.layout.setContentsMargins(0, 0, 0, 0)
        self.lbl_opacity_2 = QLabel("opacity", self)
        self.spin_opacity_2 = QDoubleSpinBox(self)
        self.spin_opacity_2.setMaximum(1.0)
        self.spin_opacity_2.setSingleStep(0.1)
        self.spin_opacity_2.setProperty("value", self.pref["spotlight.opacity"])

        spacerItem8 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        spacerItem9 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.gridLayout_3.addWidget(self.lbl_radius, 0, 0, 1, 1)
        self.gridLayout_3.addWidget(self.spin_radius, 0, 1, 1, 1)
        self.gridLayout_3.addItem(spacerItem8, 0, 2, 1, 1)
        self.gridLayout_3.addWidget(self.lbl_color_3, 1, 0, 1, 1)
        self.gridLayout_3.addWidget(self.combo_color_3, 1, 1, 1, 1)
        self.gridLayout_3.addWidget(self.lbl_opacity_2, 2, 0, 1, 1)
        self.gridLayout_3.addWidget(self.spin_opacity_2, 2, 1, 1, 1)
        self.gridLayout_3.addItem(spacerItem9, 3, 0, 1, 1)

        self.spin_radius.valueChanged.connect(lambda v: self.pref.update({"spotlight.size": round(v, 4)}))
        self.spin_opacity_2.valueChanged.connect(lambda v: self.pref.update({"spotlight.opacity": round(v, 4)}))

    # ==================================================
    def accept(self):
        self.preference.update(self.pref)
        self.setting.update(self.sett)
        self.callback()
        super().accept()

    # ==================================================
    def reject(self):
        self.preference.update(self.default)
        self.setting.update(self.default_set)
        self.callback()
        super().reject()

    # ==================================================
    def apply(self, button):
        if self.buttonBox.standardButton(button) == QDialogButtonBox.Apply:
            self.preference.update(self.pref)
            self.setting.update(self.sett)
            self.callback()



import sympy as sp
from qtpy.QtWidgets import (
    QDialog,
    QLabel,
    QTabWidget,
    QGridLayout,
    QWidget,
    QSpacerItem,
    QSizePolicy,
    QComboBox,
    QPushButton,
    QLineEdit,
    QCheckBox,
    QMessageBox,
)
from qtpy.QtCore import Qt
from gcoreutils.nsarray import NSArray
from gcoreutils.string_util import remove_space
from multipie.group.point_group import PointGroup
from multipie.group.space_group import SpaceGroup
from multipie.tag.tag_group import TagGroup
from multipie.const import __def_dict__
from qtdraw.multipie.dialog_group_info import (
    create_character_table,
    create_harmonics,
    create_harmonics_decomp,
    create_symmetry_operation,
    create_product_table,
    create_wyckoff,
    create_v_cluster,
    create_response,
    create_atomic_mp,
)
from qtdraw.multipie.setting import rcParams
from qtdraw.multipie.plot_object import (
    check_get_site,
    check_get_bond,
    check_get_site_bond,
    check_linear_combination,
    create_samb_object,
    create_combined,
    plot_equivalent_site,
    plot_equivalent_bond,
    plot_vector_equivalent_site,
    plot_orbital_equivalent_site,
    plot_site_cluster,
    plot_bond_cluster,
    plot_vector_cluster,
    plot_orbital_cluster,
)
from qtdraw.multipie.dialog_modulation import DialogModulation
from qtdraw.core.qt_logging import dprint


# ==================================================
class DialogGroup(QDialog):
    # ==================================================
    def __init__(self, core, qtdraw, width=512, height=600, parent=None):
        """
        initialize the class.

        Args:
            core (MultiPieCore): MultiPie core.
            qtdraw (QtDraw): QtDraw.
            width (int, optional): window width.
            height (int, optional): window height.
            parent (_type_, optional): Qt parent object.
        """
        super().__init__(parent)
        self._core = core
        self._qtdraw = qtdraw
        self.crystal = [{}, {}]  # space/point group.
        for crystal in __def_dict__["crystal"]:
            self.crystal[0][crystal] = [f"{i.no}. {i} ({i.IS})" for i in TagGroup.create(crystal=crystal, space_group=True)]
            self.crystal[1][crystal] = [f"{i.no}. {i} ({i.IS})" for i in TagGroup.create(crystal=crystal)]

        self.setWindowTitle(f"QtDraw - Group Operations with MultiPie Ver. {self._qtdraw._multipie_loaded}")
        self.resize(width, height)

        # default.
        tag = rcParams["group_tag"]

        self.set_group_object(tag)

        self.create_main_panel()

    # ==================================================
    def closeEvent(self, event):
        if not self.force_close:
            ret = QMessageBox.question(self, "", "Quit MultiPie Panel ?", QMessageBox.Ok, QMessageBox.Cancel)
            if ret != QMessageBox.Ok:
                event.ignore()
            else:
                self.close()
        else:
            self.close()

    # ==================================================
    def close(self):
        if self.tab2_modulated_vector_active:
            self.dialog_modulation_vector.close()
        if self.tab2_modulated_orbital_active:
            self.dialog_modulation_orbital.close()
        self._qtdraw.button_clear.show()
        self._qtdraw.button_multipie.show()
        self._qtdraw.button_repeat.show()
        super().close()

    # ==================================================
    def create_main_panel(self):
        """
        create main panel.
        """
        # group type selector.
        self.main_g_type = QComboBox(self)
        self.main_g_type.setFocusPolicy(Qt.NoFocus)
        self.main_g_type.addItems(["space group", "point group"])
        self.main_g_type.setCurrentIndex(self.pg)

        # crystal selector.
        self.main_c_type = QComboBox(self)
        self.main_c_type.setFocusPolicy(Qt.NoFocus)
        self.main_c_type.addItems(list(self.crystal[self.pg].keys()))
        self.main_c_type.setCurrentIndex(0)

        # group selector.
        self.main_group = QComboBox(self)
        self.main_group.setFocusPolicy(Qt.NoFocus)
        self.main_group.addItems(self.crystal[self.pg][self._group.tag.crystal])
        self.main_group.setCurrentIndex(0)

        # symmetry operation, character table, wyckoff, product table.
        self.main_symmetry_operation = QPushButton("symmetry operation", self)
        self.main_symmetry_operation.setFocusPolicy(Qt.NoFocus)
        self.main_character_table = QPushButton("character table", self)
        self.main_character_table.setFocusPolicy(Qt.NoFocus)
        self.main_wyckoff = QPushButton("Wyckoff position", self)
        self.main_wyckoff.setFocusPolicy(Qt.NoFocus)
        self.main_product_table = QPushButton("product table", self)
        self.main_product_table.setFocusPolicy(Qt.NoFocus)

        # harmonics.
        self.main_harmonics = QPushButton("harmonics", self)
        self.main_harmonics.setFocusPolicy(Qt.NoFocus)
        self.main_harmonics_type = QComboBox(self)
        self.main_harmonics_type.setFocusPolicy(Qt.NoFocus)
        self.main_harmonics_type.addItems(["Q", "G", "T", "M"])
        self.main_harmonics_type.setCurrentIndex(0)
        self.main_harmonics_rank_label = QLabel("rank", self)
        self.main_harmonics_rank = QComboBox(self)
        self.main_harmonics_rank.setFocusPolicy(Qt.NoFocus)
        self.main_harmonics_rank.addItems(map(str, range(12)))
        self.main_harmonics_rank.setCurrentIndex(0)
        self.main_harmonics_to_label = QLabel("irrep. decomp.", self)
        self.main_harmonics_to_pg = QComboBox(self)
        self.main_harmonics_to_pg.setFocusPolicy(Qt.NoFocus)
        self.main_harmonics_to_pg.addItems([f"{i.no}. {i}" for i in TagGroup.create()])
        self.main_harmonics_to_pg.setCurrentIndex(0)
        self.main_harmonics_to_gen = QPushButton("gen", self)
        self.main_harmonics_to_gen.setFocusPolicy(Qt.NoFocus)

        # response.
        self.main_response = QPushButton("response tensor", self)
        self.main_response.setFocusPolicy(Qt.NoFocus)
        self.main_response_rank_label = QLabel("rank", self)
        self.main_response_rank = QComboBox(self)
        self.main_response_rank.setFocusPolicy(Qt.NoFocus)
        self.main_response_rank.addItems(map(str, range(1, 5)))
        self.main_response_rank.setCurrentIndex(1)
        self.main_response_i_type = QComboBox(self)
        self.main_response_i_type.setFocusPolicy(Qt.NoFocus)
        self.main_response_i_type.addItems(["polar", "axial"])
        self.main_response_i_type.setCurrentIndex(0)
        self.main_response_t_type = QComboBox(self)
        self.main_response_t_type.setFocusPolicy(Qt.NoFocus)
        self.main_response_t_type.addItems(["E", "M"])
        self.main_response_t_type.setCurrentIndex(0)

        # atomic multipole.
        self.main_atomic_mp = QPushButton("atomic multipole", self)
        self.main_atomic_mp.setFocusPolicy(Qt.NoFocus)
        self.main_atomic_mp_head = QComboBox(self)
        self.main_atomic_mp_head.setFocusPolicy(Qt.NoFocus)
        self.main_atomic_mp_head.addItems(["", "Q", "G", "T", "M"])
        self.main_atomic_mp_head.setCurrentIndex(0)
        self.main_atomic_mp_btype = QComboBox(self)
        self.main_atomic_mp_btype.setFocusPolicy(Qt.NoFocus)
        self.main_atomic_mp_btype.addItems(["lm", "jm"])
        self.main_atomic_mp_btype.setCurrentIndex(0)
        self.main_atomic_mp_braket_label = QLabel("bra-ket", self)
        self.main_atomic_mp_bra_basis = QComboBox(self)
        self.main_atomic_mp_bra_basis.setFocusPolicy(Qt.NoFocus)
        self.main_atomic_mp_bra_basis.addItems(["s", "p", "d", "f"])
        self.main_atomic_mp_bra_basis.setCurrentIndex(0)
        self.main_atomic_mp_ket_basis = QComboBox(self)
        self.main_atomic_mp_ket_basis.setFocusPolicy(Qt.NoFocus)
        self.main_atomic_mp_ket_basis.addItems(["s", "p", "d", "f"])
        self.main_atomic_mp_ket_basis.setCurrentIndex(0)

        # virtual cluster.
        self.main_v_cluster = QPushButton("virtual cluster", self)
        self.main_v_cluster.setFocusPolicy(Qt.NoFocus)
        self.main_v_cluster_wp = QComboBox(self)
        self.main_v_cluster_wp.setFocusPolicy(Qt.NoFocus)
        self.main_v_cluster_wp.addItems(self.wp)
        self.main_v_cluster_wp.setCurrentIndex(0)
        self.main_v_cluster_bond = QLineEdit("1", self)

        # product irrep. decomposition.
        self.main_irrep_sym_label = QLabel("symmetric", self)
        self.main_irrep_sym_label.setFocusPolicy(Qt.NoFocus)
        self.main_irrep_asym_label = QLabel("anti-sym.", self)
        self.main_irrep_asym_label.setFocusPolicy(Qt.NoFocus)
        self.main_irrep1 = QComboBox(self)
        self.main_irrep1.setFocusPolicy(Qt.NoFocus)
        self.main_irrep1.addItems(self.irrep)
        self.main_irrep1.setCurrentIndex(0)
        self.main_irrep2 = QComboBox(self)
        self.main_irrep2.setFocusPolicy(Qt.NoFocus)
        self.main_irrep2.addItems(self.irrep)
        self.main_irrep2.setCurrentIndex(0)
        self.main_irrep_sym = QLabel("", self)
        self.main_irrep_sym.setFocusPolicy(Qt.NoFocus)
        self.main_irrep = QComboBox(self)
        self.main_irrep.setFocusPolicy(Qt.NoFocus)
        self.main_irrep.addItems(self.irrep)
        self.main_irrep.setCurrentIndex(0)
        self.main_irrep_asym = QLabel("", self)
        self.main_irrep_asym.setFocusPolicy(Qt.NoFocus)
        self.set_product_irrep("")

        # clear button.
        self.main_clear = QPushButton("clear")
        self.main_clear.setFocusPolicy(Qt.NoFocus)

        # site bond description.
        self.main_site_bond_label = QLabel("SITE: [x,y,z],    BOND: [tail];[head] / [vector]@[center] / [start]:[vector]", self)
        self.main_site_bond_label.setFocusPolicy(Qt.NoFocus)

        # tab contents.
        self.tab = QTabWidget(self)
        self.create_tab1_panel()
        self.create_tab2_panel()

        # main layout
        self.layout = QGridLayout(self)
        self.layout.addWidget(self.main_g_type, 0, 0, 1, 2)
        self.layout.addWidget(self.main_c_type, 0, 2, 1, 3)
        self.layout.addWidget(self.main_group, 0, 5, 1, 4)

        self.layout.addWidget(self.main_symmetry_operation, 1, 0, 1, 2)
        self.layout.addWidget(self.main_character_table, 1, 2, 1, 3)
        self.layout.addWidget(self.main_wyckoff, 1, 5, 1, 2)
        self.layout.addWidget(self.main_product_table, 1, 7, 1, 2)

        self.layout.addWidget(self.main_irrep_sym_label, 2, 0, 1, 1)
        self.layout.addWidget(self.main_irrep1, 2, 1, 1, 1)
        self.layout.addWidget(self.main_irrep2, 2, 2, 1, 2)
        self.layout.addWidget(self.main_irrep_sym, 2, 4, 1, 2)
        self.layout.addWidget(self.main_irrep_asym_label, 2, 6, 1, 1)
        self.layout.addWidget(self.main_irrep, 2, 7, 1, 2)
        self.layout.addWidget(self.main_irrep_asym, 2, 9, 1, 1)

        self.layout.addWidget(self.main_harmonics, 3, 0, 1, 2)
        self.layout.addWidget(self.main_harmonics_type, 3, 2, 1, 1)
        self.layout.addWidget(self.main_harmonics_rank_label, 3, 3, 1, 1)
        self.layout.addWidget(self.main_harmonics_rank, 3, 4, 1, 1)
        self.layout.addWidget(self.main_harmonics_to_label, 3, 6, 1, 1)
        self.layout.addWidget(self.main_harmonics_to_pg, 3, 7, 1, 1)
        self.layout.addWidget(self.main_harmonics_to_gen, 3, 8, 1, 1)

        self.layout.addWidget(self.main_response, 4, 0, 1, 2)
        self.layout.addWidget(self.main_response_t_type, 4, 2, 1, 1)
        self.layout.addWidget(self.main_response_i_type, 4, 3, 1, 1)
        self.layout.addWidget(self.main_response_rank_label, 4, 4, 1, 1)
        self.layout.addWidget(self.main_response_rank, 4, 5, 1, 1)

        self.layout.addWidget(self.main_atomic_mp, 5, 0, 1, 2)
        self.layout.addWidget(self.main_atomic_mp_head, 5, 2, 1, 1)
        self.layout.addWidget(self.main_atomic_mp_btype, 5, 3, 1, 1)
        self.layout.addWidget(self.main_atomic_mp_braket_label, 5, 4, 1, 1)
        self.layout.addWidget(self.main_atomic_mp_bra_basis, 5, 5, 1, 2)
        self.layout.addWidget(self.main_atomic_mp_ket_basis, 5, 7, 1, 2)

        self.layout.addWidget(self.main_v_cluster, 6, 0, 1, 2)
        self.layout.addWidget(self.main_v_cluster_wp, 6, 2, 1, 2)
        self.layout.addWidget(self.main_v_cluster_bond, 6, 4, 1, 2)

        self.layout.addWidget(self.main_clear, 7, 9, 1, 1)

        self.layout.addWidget(self.main_site_bond_label, 8, 0, 1, 9)

        self.layout.addWidget(self.tab, 9, 0, 1, 10)

        # connections.
        self.main_g_type.currentIndexChanged.connect(self.set_group_type)
        self.main_c_type.currentTextChanged.connect(self.set_crystal_type)
        self.main_group.currentTextChanged.connect(self.set_group)

        self.main_symmetry_operation.clicked.connect(self.show_symmetry_operation)
        self.main_character_table.clicked.connect(self.show_character_table)
        self.main_wyckoff.clicked.connect(self.show_wyckoff)
        self.main_product_table.clicked.connect(self.show_product_table)

        self.main_harmonics.clicked.connect(self.show_harmonics)
        self.main_harmonics_to_gen.clicked.connect(self.show_harmonics_decomp)

        self.main_response.clicked.connect(self.show_response)

        self.main_atomic_mp.clicked.connect(self.show_atomic_mp)
        self.main_atomic_mp_btype.currentIndexChanged.connect(self.set_atomic_mp_btype)

        self.main_v_cluster.clicked.connect(self.show_v_cluster)

        self.main_irrep1.currentTextChanged.connect(self.set_product_irrep)
        self.main_irrep2.currentTextChanged.connect(self.set_product_irrep)
        self.main_irrep.currentTextChanged.connect(self.set_product_irrep)

        self.main_clear.clicked.connect(self.clear_data)

    # ==================================================
    def set_group_object(self, tag):
        """
        set group object and properties.

        Args:
            tag (TagGroup or str): group tag.

        Notes:
            - set the following properties, self.pg, self._group, self._pgroup, self.wp, self.irrep, self.n_op, self.n_pset, self.pset.
        """
        if type(tag) == str:
            if tag.count(" ") > 0:
                tag = tag.split(" ")[1]
            tag = TagGroup(tag)
        self.pg = tag.is_point_group()
        if self.pg:
            self._group = PointGroup(tag, self._core)
            self._pgroup = self._group
            self._qtdraw.setting["cluster"] = True
            self._qtdraw._toggle_clip(False)
            self._qtdraw.button_repeat.hide()
            self.n_pset = 1
            self.pset = NSArray("{[0,0,0]}")
        else:
            self._group = SpaceGroup(tag, self._core)
            self._pgroup = self._group.pg
            self._qtdraw.setting["cluster"] = False
            self._qtdraw._toggle_clip(True)
            self._qtdraw.button_repeat.show()
            self.n_pset = len(self._group.symmetry_operation.plus_set)
            self.pset = self._group.symmetry_operation.plus_set
        self.wp = list(map(str, self._pgroup.wyckoff.keys()))[::-1]
        self.irrep = list(map(str, self._pgroup.character.irrep_list))
        self.n_op = len(self._group.symmetry_operation)

    # ==================================================
    def set_group(self, tag):
        """
        set current group.

        Args:
            tag (TagGroup or str): group tag.
        """
        if tag == "":
            return

        self.set_group_object(tag)

        self.main_v_cluster_wp.clear()
        self.main_v_cluster_wp.addItems(self.wp)
        self.main_v_cluster_wp.setCurrentIndex(0)

        self.main_irrep1.clear()
        self.main_irrep1.addItems(self.irrep)
        self.main_irrep1.setCurrentIndex(0)
        self.main_irrep2.clear()
        self.main_irrep2.addItems(self.irrep)
        self.main_irrep2.setCurrentIndex(0)
        self.main_irrep.clear()
        self.main_irrep.addItems(self.irrep)
        self.main_irrep.setCurrentIndex(0)

        self.tab1_pgharm_type.setCurrentIndex(0)
        self.tab1_pgharm_rank.setCurrentIndex(0)
        self.tab1_pgharm_rank_select()

    # ==================================================
    def set_group_type(self, is_point_group):
        """
        set group type (point or space).

        Args:
            is_point_group (bool): point group ?
        """
        if is_point_group:
            tag = str(self._pgroup.tag)
        else:
            tag = self.main_group.currentText().split(" ")[1]
            if tag[-2:] == "-1":
                tag = tag[:-2]
            tag = tag + "^1"

        self.set_group_object(tag)

        crystal = self.main_c_type.currentText()
        self.main_group.clear()
        lst = self.crystal[self.pg][crystal]
        no = lst.index(str(self._group.tag.no) + ". " + tag + " (" + self._group.tag.IS + ")")
        self.main_group.addItems(lst)
        self.main_group.setCurrentIndex(no)

    # ==================================================
    def set_crystal_type(self, crystal):
        """
        set crystal type.

        Args:
            crystal (str): crystal type.
        """
        tags = self.crystal[self.pg][crystal]
        self.set_group_object(tags[0])

        self.main_group.clear()
        self.main_group.addItems(tags)
        self.main_group.setCurrentIndex(0)
        self._qtdraw.set_crystal(crystal)

    # ==================================================
    def set_product_irrep(self, _):
        """
        set product irrep decomposition.
        """
        irrep1 = self.main_irrep1.currentText()
        if irrep1 not in self.irrep:
            return
        irrep2 = self.main_irrep2.currentText()
        if irrep2 not in self.irrep:
            return
        irrep = self.main_irrep.currentText()
        if irrep not in self.irrep:
            return

        def remove_latex(s):
            s = (
                s.replace("_{", "")
                .replace("}", "")
                .replace("^{", "")
                .replace(r"\prime", "'")
                .replace("(", "")
                .replace(")", "")
                .replace("0", "-")
            )
            return s

        sym = self._pgroup.character.symmetric_product_decomposition((irrep1, irrep2), ret_ex=True)
        asym = self._pgroup.character.anti_symmetric_product_decomposition(irrep, ret_ex=True)
        self.main_irrep_sym.setText(remove_latex(str(sym)))
        self.main_irrep_asym.setText(remove_latex(str(asym)))

    # ==================================================
    def show_symmetry_operation(self):
        """
        show symmetry operation panel.
        """
        create_symmetry_operation(self._group, self)

    # ==================================================
    def show_character_table(self):
        """
        show character table panel.
        """
        create_character_table(self._pgroup, self)

    # ==================================================
    def show_wyckoff(self):
        """
        show wyckoff panel.
        """
        create_wyckoff(self._group, self)

    # ==================================================
    def show_product_table(self):
        """
        show product table panel.
        """
        create_product_table(self._pgroup, self)

    # ==================================================
    def show_harmonics(self):
        """
        show harmonics panel.
        """
        head = self.main_harmonics_type.currentText()
        rank = int(self.main_harmonics_rank.currentText())
        create_harmonics(self._pgroup, rank, head, self._qtdraw, self)

    # ==================================================
    def show_harmonics_decomp(self):
        """
        show harmonics decomposition panel.
        """
        head = self.main_harmonics_type.currentText()
        rank = int(self.main_harmonics_rank.currentText())
        to_pg = self.main_harmonics_to_pg.currentText()
        create_harmonics_decomp(self._pgroup, rank, head, to_pg, self)

    # ==================================================
    def show_response(self):
        """
        show response panel.
        """
        rank = int(self.main_response_rank.currentText())
        i_type = self.main_response_i_type.currentText()
        t_type = self.main_response_t_type.currentText()
        create_response(self._pgroup, rank, i_type, t_type, self)

    # ==================================================
    def show_atomic_mp(self):
        """
        show atomic multipole panel.
        """
        head = self.main_atomic_mp_head.currentText()
        btype = self.main_atomic_mp_btype.currentText()
        bra = self.main_atomic_mp_bra_basis.currentText()
        ket = self.main_atomic_mp_ket_basis.currentText()

        spinful = btype == "jm"
        am = self._pgroup.atomic_multipole_basis(bra, ket, spinful)
        if head != "":
            am = am.select(head=head)

        create_atomic_mp(self._pgroup, bra, ket, am, self)

    # ==================================================
    def set_atomic_mp_btype(self, g):
        """
        set atomic multipole basis type.

        Args:
            g (int): spinless or not.
        """
        if g == 0:
            lst = ["s", "p", "d", "f"]
        else:  # jm
            lst = ["(1/2,0)", "(1/2,1)", "(3/2,1)", "(3/2,2)", "(5/2,2)", "(5/2,3)", "(7/2,3)"]
        self.main_atomic_mp_bra_basis.clear()
        self.main_atomic_mp_ket_basis.clear()
        self.main_atomic_mp_bra_basis.addItems(lst)
        self.main_atomic_mp_ket_basis.addItems(lst)
        self.main_atomic_mp_bra_basis.setCurrentIndex(0)
        self.main_atomic_mp_ket_basis.setCurrentIndex(0)

    # ==================================================
    def show_v_cluster(self):
        """
        show virtual cluster panel.
        """
        wp = self.main_v_cluster_wp.currentText()
        bond = self.main_v_cluster_bond.text()
        create_v_cluster(self._pgroup, wp, bond, self._qtdraw, self)

    # ==================================================
    def clear_data(self):
        """
        clear all data.
        """
        self._qtdraw._clear()
        self._qtdraw.set_crystal(self.main_c_type.currentText())

    # ==================================================
    def create_tab1_panel(self):
        """
        create tab1 panel.
        """
        self.tab1 = QWidget()
        self.tab.addTab(self.tab1, "object drawing")

        self.tab1_site_label = QLabel("SITE: draw equivalent sites.\n 1. input representative SITE, + ENTER.", self)
        self.tab1_site_pos = QLineEdit("[ 1/2, 1/2, 0 ]", self)

        self.tab1_bond_label = QLabel(
            "BOND: draw equivalent bonds.\n 1. input representative BOND, + ENTER.",
            self,
        )
        self.tab1_bond_pos = QLineEdit("[ 0, 0, 0 ] ; [ 1/2, 1/2, 0 ]", self)

        self.tab1_vector_label = QLabel(
            "VECTOR: draw vectors at equivalent sites or bonds.\n 1. choose type, 2. input vector [x,y,z] # representative SITE/BOND, + ENTER.",
            self,
        )
        self.tab1_vector_pos = QLineEdit("[ 0, 0, 1 ] # [ 1/2, 1/2, 0 ]", self)
        self.tab1_vector_type = QComboBox(self)
        self.tab1_vector_type.addItems(["Q", "G", "T", "M"])
        self.tab1_vector_type.setCurrentIndex(0)

        self.tab1_orbital_label = QLabel(
            "ORBITAL: draw orbitals at equivalent sites or bonds.\n 1. choose type, 2. input orbital (xyz polynomial) # representative SITE/BOND, + ENTER.",
            self,
        )
        self.tab1_orbital_pos = QLineEdit("3z**2 - r**2 # [ 1/4, 1/4, 0 ]", self)
        self.tab1_orbital_type = QComboBox(self)
        self.tab1_orbital_type.addItems(["Q", "G", "T", "M"])
        self.tab1_orbital_type.setCurrentIndex(0)

        self.tab1_pgharm_label = QLabel(
            "POINT-GROUP HARMONICS: draw point-group harmonics at equivalent sites or bonds.\n 1. choose (type,rank,irrep.), 2. input representative SITE/BOND, + ENTER.\n \u21d2  used expression is shown (in LaTeX form).",
            self,
        )
        self.tab1_pgharm_pos = QLineEdit("[ 0, 0, 0 ]", self)
        self.tab1_pgharm_type = QComboBox(self)
        self.tab1_pgharm_type.addItems(["Q", "G", "T", "M"])
        self.tab1_pgharm_type.setCurrentIndex(0)
        self.tab1_pgharm_rank = QComboBox(self)
        self.tab1_pgharm_rank.setFocusPolicy(Qt.NoFocus)
        self.tab1_pgharm_rank.addItems(map(str, range(12)))
        self.tab1_pgharm_rank.setCurrentIndex(0)
        self.tab1_pgharm_irrep = QComboBox(self)
        self.tab1_pgharm_irrep.setFocusPolicy(Qt.NoFocus)
        hs0 = self._pgroup.harmonics.select(rank=0, head="Q")
        self.tab1_pgharm_irrep.addItems(["Q" + str(i)[2:] for i in hs0])
        self.tab1_pgharm_irrep.setCurrentIndex(0)
        self.tab1_pgharm_exp_label = QLabel("expression", self)
        self.tab1_pgharm_exp = QLineEdit("1", self)
        self.tab1_pgharm_exp_latex = QCheckBox("LaTeX", self)
        self.tab1_pgharm_exp_latex.setChecked(False)

        self.tab1_v_spacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.tab1_grid = QGridLayout(self.tab1)
        self.tab1_grid.addWidget(self.tab1_site_label, 0, 0, 1, 9)
        self.tab1_grid.addWidget(self.tab1_site_pos, 1, 0, 1, 9)

        self.tab1_grid.addWidget(self.tab1_bond_label, 2, 0, 1, 9)
        self.tab1_grid.addWidget(self.tab1_bond_pos, 3, 0, 1, 9)

        self.tab1_grid.addWidget(self.tab1_vector_label, 4, 0, 1, 9)
        self.tab1_grid.addWidget(self.tab1_vector_type, 5, 0, 1, 1)
        self.tab1_grid.addWidget(self.tab1_vector_pos, 5, 1, 1, 8)

        self.tab1_grid.addWidget(self.tab1_orbital_label, 6, 0, 1, 9)
        self.tab1_grid.addWidget(self.tab1_orbital_type, 7, 0, 1, 1)
        self.tab1_grid.addWidget(self.tab1_orbital_pos, 7, 1, 1, 8)

        self.tab1_grid.addWidget(self.tab1_pgharm_label, 8, 0, 1, 9)
        self.tab1_grid.addWidget(self.tab1_pgharm_type, 9, 0, 1, 1)
        self.tab1_grid.addWidget(self.tab1_pgharm_rank, 9, 1, 1, 1)
        self.tab1_grid.addWidget(self.tab1_pgharm_irrep, 9, 2, 1, 2)
        self.tab1_grid.addWidget(self.tab1_pgharm_pos, 9, 4, 1, 5)
        self.tab1_grid.addWidget(self.tab1_pgharm_exp_label, 10, 0, 1, 1)
        self.tab1_grid.addWidget(self.tab1_pgharm_exp, 10, 1, 1, 7)
        self.tab1_grid.addWidget(self.tab1_pgharm_exp_latex, 10, 8, 1, 1)

        self.tab1_grid.addItem(self.tab1_v_spacer, 11, 1, 1, 1)

        # connections.
        self.tab1_site_pos.returnPressed.connect(self.tab1_plot_site)
        self.tab1_bond_pos.returnPressed.connect(self.tab1_plot_bond)
        self.tab1_vector_pos.returnPressed.connect(self.tab1_plot_vector)
        self.tab1_orbital_pos.returnPressed.connect(self.tab1_plot_orbital)
        self.tab1_pgharm_type.currentIndexChanged.connect(self.tab1_pgharm_rank_select)
        self.tab1_pgharm_rank.currentIndexChanged.connect(self.tab1_pgharm_rank_select)
        self.tab1_pgharm_irrep.currentIndexChanged.connect(self.tab1_select_pg_harmonics)
        self.tab1_pgharm_type.currentIndexChanged.connect(self.tab1_select_pg_harmonics)
        self.tab1_pgharm_rank.currentIndexChanged.connect(self.tab1_select_pg_harmonics)
        self.tab1_pgharm_exp_latex.toggled.connect(self.tab1_select_pg_harmonics)
        self.tab1_pgharm_pos.returnPressed.connect(self.tab1_plot_pg_harmonics)

    # ==================================================
    def tab1_plot_site(self):
        """
        plot equivalent sites.
        """
        r_site = check_get_site(self.tab1_site_pos.text())
        if r_site is None:
            return

        if self.pg:
            site = self._group.site_mapping(r_site)
        else:
            site = self._group.site_mapping(r_site, plus_set=True)

        plot_equivalent_site(self._qtdraw, site, self.n_pset)

    # ==================================================
    def tab1_plot_bond(self):
        """
        plot equivalent bonds.
        """
        r_bond = check_get_bond(self.tab1_bond_pos.text())
        if r_bond is None:
            return

        if self.pg:
            bond, nondirectional = self._group.bond_mapping(r_bond)
        else:
            bond, nondirectional = self._group.bond_mapping(r_bond, plus_set=True)

        plot_equivalent_bond(self._qtdraw, bond, nondirectional, self.n_pset)

    # ==================================================
    def tab1_plot_vector(self):
        """
        plot vectors at equivalent sites/bonds.
        """
        head = self.tab1_vector_type.currentText()
        txt = self.tab1_vector_pos.text().split("#")
        if len(txt) != 2:
            return
        vector, r_site_bond = txt
        vector = check_get_site(vector)
        if vector is None:
            return
        r_site = check_get_site_bond(r_site_bond)
        if r_site is None:
            return

        if self.pg:
            site = self._group.site_mapping(r_site)
        else:
            site = self._group.site_mapping(r_site, plus_set=True)
        site = NSArray.from_str(site.keys())

        plot_vector_equivalent_site(self._qtdraw, site, vector, head, self.n_pset)

    # ==================================================
    def tab1_plot_orbital(self):
        """
        plot orbitals at equivalent sites/bonds.
        """
        head = self.tab1_orbital_type.currentText()
        txt = self.tab1_orbital_pos.text().split("#")
        if len(txt) != 2:
            return
        orbital, r_site_bond = txt
        r_site = check_get_site_bond(r_site_bond)
        if r_site is None:
            return

        if self.pg:
            site = self._group.site_mapping(r_site)
        else:
            site = self._group.site_mapping(r_site, plus_set=True)
        site = NSArray.from_str(site.keys())

        plot_orbital_equivalent_site(self._qtdraw, site, orbital, head, self.n_pset)

    # ==================================================
    def tab1_select_pg_harmonics(self):
        """
        select poing-group harmonics.
        """
        head = self.tab1_pgharm_type.currentText()
        head1 = head.replace("T", "Q").replace("M", "G")
        rank = int(self.tab1_pgharm_rank.currentText())
        hs = self._pgroup.harmonics.select(rank=rank, head=head1)
        h = hs[self.tab1_pgharm_irrep.currentIndex()]
        ex = h.expression(v=NSArray.vector3d("Q"))
        if self.tab1_pgharm_exp_latex.checkState():
            ex = ex.latex()
        self.tab1_pgharm_exp.setText(str(ex))

    # ==================================================
    def tab1_pgharm_rank_select(self):
        """
        select point-group harmonics rank.
        """
        head = self.tab1_pgharm_type.currentText()
        head1 = head.replace("T", "Q").replace("M", "G")
        rank = int(self.tab1_pgharm_rank.currentText())
        hs = self._pgroup.harmonics.select(rank=rank, head=head1)
        self.tab1_pgharm_irrep.clear()
        comb = [head + str(i)[2:] for i in hs]
        self.tab1_pgharm_irrep.addItems(comb)
        self.tab1_pgharm_irrep.setCurrentIndex(0)
        self.tab1_pgharm_exp.setText("1")

    # ==================================================
    def tab1_plot_pg_harmonics(self):
        """
        plot poing-group harmonics at equivalent sites/bonds.
        """
        head = self.tab1_pgharm_type.currentText()
        rank = int(self.tab1_pgharm_rank.currentText())
        harmonics_set = self._pgroup.harmonics.select(rank=rank, head=head.replace("T", "Q").replace("M", "G"))
        harmonics = harmonics_set[self.tab1_pgharm_irrep.currentIndex()]
        orbital = harmonics.expression(v=NSArray.vector3d("Q"))

        r_site = check_get_site_bond(self.tab1_pgharm_pos.text())
        if r_site is None:
            return

        if self.pg:
            site = self._group.site_mapping(r_site)
        else:
            site = self._group.site_mapping(r_site, plus_set=True)
        site = NSArray.from_str(site.keys())

        plot_orbital_equivalent_site(self._qtdraw, site, orbital, head, self.n_pset)

    # ==================================================
    def create_tab2_panel(self):
        """
        create tab2 panel.
        """
        self.tab2 = QWidget()
        self.tab.addTab(self.tab2, "basis drawing")

        self.tab2_site_proj_label = QLabel(
            "SITE: draw site-cluster basis.\n 1. input representative SITE, + ENTER,\n \u21d2  2. choose basis, 3. push `draw`.",
            self,
        )
        self.tab2_site_proj_pos = QLineEdit("[ 1/2, 1/2, 0 ]", self)
        self.tab2_site_proj_label1 = QLabel("\u21d2", self)
        self.tab2_site_proj_label1.setAlignment(Qt.AlignCenter)
        self.tab2_site_proj_irrep1 = QComboBox(self)
        self.tab2_site_proj_irrep1.addItems([""])
        self.tab2_site_proj_irrep1.setCurrentIndex(0)
        self.tab2_site_proj_draw_button = QPushButton("draw", self)
        self.tab2_site_proj_draw_button.setFocusPolicy(Qt.NoFocus)

        self.tab2_bond_proj_label = QLabel(
            "BOND: draw bond-cluster basis.\n 1. input representative BOND, + ENTER,\n \u21d2  2. choose basis, 3. push `draw`.",
            self,
        )
        self.tab2_bond_proj_pos = QLineEdit("[ 1/2, 1/2, 0 ] @ [ 1/4, 1/4, 0 ]", self)
        self.tab2_bond_proj_label1 = QLabel("\u21d2", self)
        self.tab2_bond_proj_label1.setAlignment(Qt.AlignCenter)
        self.tab2_bond_proj_irrep1 = QComboBox(self)
        self.tab2_bond_proj_irrep1.addItems([""])
        self.tab2_bond_proj_irrep1.setCurrentIndex(0)
        self.tab2_bond_proj_draw_button = QPushButton("draw", self)
        self.tab2_bond_proj_draw_button.setFocusPolicy(Qt.NoFocus)

        self.tab2_vector_proj_label = QLabel(
            "VECTOR: draw symmetry-adapted vector.\n 1. choose type, 2. input representative SITE/BOND, + ENTER,\n \u21d2  3. choose (type,basis), 4. push `draw` or 3. input linear combination (LC), + ENTER or 3. push `modulation`.",
            self,
        )
        self.tab2_vector_proj_pos = QLineEdit("[ 1/2, 1/2, 0 ]", self)
        self.tab2_vector_proj_type = QComboBox(self)
        self.tab2_vector_proj_type.addItems(["Q", "G", "T", "M"])
        self.tab2_vector_proj_type.setCurrentIndex(0)
        self.tab2_vector_proj_type1 = QComboBox(self)
        self.tab2_vector_proj_type1.addItems(["Q", "G", "T", "M"])
        self.tab2_vector_proj_type1.setCurrentIndex(0)
        self.tab2_vector_proj_label1 = QLabel("\u21d2", self)
        self.tab2_vector_proj_label1.setAlignment(Qt.AlignCenter)
        self.tab2_vector_proj_irrep1 = QComboBox(self)
        self.tab2_vector_proj_irrep1.addItems([""])
        self.tab2_vector_proj_irrep1.setCurrentIndex(0)
        self.tab2_vector_proj_draw_button = QPushButton("draw", self)
        self.tab2_vector_proj_draw_button.setFocusPolicy(Qt.NoFocus)
        self.tab2_vector_proj_lc_label = QLabel("LC", self)
        self.tab2_vector_proj_lc_label.setAlignment(Qt.AlignCenter)
        self.tab2_vector_proj_lc = QLineEdit("(Q01+Q02)/sqrt(2)", self)
        self.tab2_vector_modulation_button = QPushButton("modulation", self)
        self.tab2_vector_modulation_button.setFocusPolicy(Qt.NoFocus)
        self.tab2_vector_modulation_type = QComboBox(self)
        self.tab2_vector_modulation_type.addItems(["Q,G", "T,M"])
        self.tab2_vector_modulation_type.setCurrentIndex(0)
        self.tab2_vector_modulation_list = QLineEdit("", self)

        self.tab2_orbital_proj_label = QLabel(
            "ORBITAL draw symmetry-adapted orbital.\n 1. choose (type,rank), 2. input representative SITE/BOND, + ENTER,\n \u21d2  3. choose (type,basis), 4. push `draw` or 3. input linear combination (LC), + ENTER or 3. push `modulation`.",
            self,
        )
        self.tab2_orbital_proj_pos = QLineEdit("[ 0, 0, 0 ]; [ 1/2, 1/2, 0 ]", self)
        self.tab2_orbital_proj_type = QComboBox(self)
        self.tab2_orbital_proj_type.addItems(["Q", "G", "T", "M"])
        self.tab2_orbital_proj_type.setCurrentIndex(0)
        self.tab2_orbital_proj_type1 = QComboBox(self)
        self.tab2_orbital_proj_type1.addItems(["Q", "G", "T", "M"])
        self.tab2_orbital_proj_type1.setCurrentIndex(0)
        self.tab2_orbital_proj_label1 = QLabel("\u21d2", self)
        self.tab2_orbital_proj_label1.setAlignment(Qt.AlignCenter)
        self.tab2_orbital_proj_irrep1 = QComboBox(self)
        self.tab2_orbital_proj_irrep1.addItems([""])
        self.tab2_orbital_proj_irrep1.setCurrentIndex(0)
        self.tab2_orbital_proj_rank = QComboBox(self)
        self.tab2_orbital_proj_rank.setFocusPolicy(Qt.NoFocus)
        self.tab2_orbital_proj_rank.addItems(map(str, range(12)))
        self.tab2_orbital_proj_rank.setCurrentIndex(1)
        self.tab2_orbital_proj_draw_button = QPushButton("draw", self)
        self.tab2_orbital_proj_draw_button.setFocusPolicy(Qt.NoFocus)
        self.tab2_orbital_proj_lc_label = QLabel("LC", self)
        self.tab2_orbital_proj_lc_label.setAlignment(Qt.AlignCenter)
        self.tab2_orbital_proj_lc = QLineEdit("(Q01+Q02)/sqrt(2)", self)
        self.tab2_orbital_modulation_button = QPushButton("modulation", self)
        self.tab2_orbital_modulation_button.setFocusPolicy(Qt.NoFocus)
        self.tab2_orbital_modulation_type = QComboBox(self)
        self.tab2_orbital_modulation_type.addItems(["Q,G", "T,M"])
        self.tab2_orbital_modulation_type.setCurrentIndex(0)
        self.tab2_orbital_modulation_list = QLineEdit("", self)

        self.tab2_hopping_proj_label = QLabel(
            "HOPPING: draw hopping direction.\n 1. input representative BOND, + ENTER.",
            self,
        )
        self.tab2_hopping_proj_pos = QLineEdit("[ 0, 0, 0 ] ; [ 1/2, 1/2, 0 ]", self)

        self.tab2_vspacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.tab2_grid = QGridLayout(self.tab2)

        self.tab2_grid.addWidget(self.tab2_site_proj_label, 0, 0, 1, 9)
        self.tab2_grid.addWidget(self.tab2_site_proj_pos, 1, 0, 1, 9)
        self.tab2_grid.addWidget(self.tab2_site_proj_label1, 2, 0, 1, 1)
        self.tab2_grid.addWidget(self.tab2_site_proj_irrep1, 2, 1, 1, 7)
        self.tab2_grid.addWidget(self.tab2_site_proj_draw_button, 2, 8, 1, 1)

        self.tab2_grid.addWidget(self.tab2_bond_proj_label, 3, 0, 1, 9)
        self.tab2_grid.addWidget(self.tab2_bond_proj_pos, 4, 0, 1, 9)
        self.tab2_grid.addWidget(self.tab2_bond_proj_label1, 5, 0, 1, 1)
        self.tab2_grid.addWidget(self.tab2_bond_proj_irrep1, 5, 1, 1, 7)
        self.tab2_grid.addWidget(self.tab2_bond_proj_draw_button, 5, 8, 1, 1)

        self.tab2_grid.addWidget(self.tab2_vector_proj_label, 6, 0, 1, 9)
        self.tab2_grid.addWidget(self.tab2_vector_proj_type, 7, 0, 1, 1)
        self.tab2_grid.addWidget(self.tab2_vector_proj_pos, 7, 1, 1, 8)
        self.tab2_grid.addWidget(self.tab2_vector_proj_label1, 8, 0, 1, 1)
        self.tab2_grid.addWidget(self.tab2_vector_proj_type1, 8, 1, 1, 1)
        self.tab2_grid.addWidget(self.tab2_vector_proj_irrep1, 8, 2, 1, 6)
        self.tab2_grid.addWidget(self.tab2_vector_proj_draw_button, 8, 8, 1, 1)
        self.tab2_grid.addWidget(self.tab2_vector_proj_lc_label, 9, 0, 1, 1)
        self.tab2_grid.addWidget(self.tab2_vector_proj_lc, 9, 1, 1, 8)
        self.tab2_grid.addWidget(self.tab2_vector_modulation_button, 10, 0, 1, 2)
        self.tab2_grid.addWidget(self.tab2_vector_modulation_type, 10, 2, 1, 1)
        self.tab2_grid.addWidget(self.tab2_vector_modulation_list, 10, 3, 1, 6)

        self.tab2_grid.addWidget(self.tab2_orbital_proj_label, 11, 0, 1, 9)
        self.tab2_grid.addWidget(self.tab2_orbital_proj_type, 12, 0, 1, 1)
        self.tab2_grid.addWidget(self.tab2_orbital_proj_rank, 12, 1, 1, 1)
        self.tab2_grid.addWidget(self.tab2_orbital_proj_pos, 12, 2, 1, 7)
        self.tab2_grid.addWidget(self.tab2_orbital_proj_label1, 13, 0, 1, 1)
        self.tab2_grid.addWidget(self.tab2_orbital_proj_type1, 13, 1, 1, 1)
        self.tab2_grid.addWidget(self.tab2_orbital_proj_irrep1, 13, 2, 1, 6)
        self.tab2_grid.addWidget(self.tab2_orbital_proj_draw_button, 13, 8, 1, 1)
        self.tab2_grid.addWidget(self.tab2_orbital_proj_lc_label, 14, 0, 1, 1)
        self.tab2_grid.addWidget(self.tab2_orbital_proj_lc, 14, 1, 1, 8)
        self.tab2_grid.addWidget(self.tab2_orbital_modulation_button, 15, 0, 1, 2)
        self.tab2_grid.addWidget(self.tab2_orbital_modulation_type, 15, 2, 1, 1)
        self.tab2_grid.addWidget(self.tab2_orbital_modulation_list, 15, 3, 1, 6)

        self.tab2_grid.addWidget(self.tab2_hopping_proj_label, 16, 0, 1, 9)
        self.tab2_grid.addWidget(self.tab2_hopping_proj_pos, 17, 0, 1, 9)

        self.tab2_grid.addItem(self.tab2_vspacer, 18, 1, 1, 1)

        # connections.
        self.tab2_site_proj_pos.returnPressed.connect(self.tab2_gen_site_cluster)
        self.tab2_site_proj_draw_button.clicked.connect(self.tab2_plot_site_cluster_samb)
        self.tab2_bond_proj_pos.returnPressed.connect(self.tab2_gen_bond_cluster)
        self.tab2_bond_proj_draw_button.clicked.connect(self.tab2_plot_bond_cluster_samb)
        self.tab2_vector_proj_pos.returnPressed.connect(self.tab2_gen_z_samb_vector)
        self.tab2_vector_proj_type1.currentIndexChanged.connect(self.tab2_select_z_samb_vector)
        self.tab2_vector_proj_draw_button.clicked.connect(self.tab2_plot_z_samb_vector)
        self.tab2_vector_proj_lc.returnPressed.connect(self.tab2_plot_z_samb_vector_lc)
        self.tab2_vector_modulation_button.clicked.connect(self.tab2_create_modulated_vector)
        self.tab2_orbital_proj_pos.returnPressed.connect(self.tab2_gen_z_samb_orbital)
        self.tab2_orbital_proj_type1.currentIndexChanged.connect(self.tab2_select_z_samb_orbital)
        self.tab2_orbital_proj_draw_button.clicked.connect(self.tab2_plot_z_samb_orbital)
        self.tab2_orbital_proj_lc.returnPressed.connect(self.tab2_plot_z_samb_orbital_lc)
        self.tab2_orbital_modulation_button.clicked.connect(self.tab2_create_modulated_orbital)
        self.tab2_hopping_proj_pos.returnPressed.connect(self.tab2_plot_z_samb_hopping)

        self.tab2_modulated_vector_active = False
        self.tab2_modulated_orbital_active = False

    # ==================================================
    def tab2_gen_site_cluster(self):
        """
        generate site cluster SAMB.
        """
        pos = check_get_site(self.tab2_site_proj_pos.text())
        if pos is None:
            return

        combined_info = create_combined(self._group, self._pgroup, pos, 0, "Q")

        self.tab2_site_proj_c_samb, self.tab2_site_proj_site, self.tab2_site_proj_z_samb = combined_info

        self.tab2_site_cluster_samb_select()

    # ==================================================
    def tab2_site_cluster_samb_select(self):
        """
        create site cluster SAMB select list.
        """
        self.tab2_site_proj_comb_select = self.tab2_site_proj_z_samb["Q"]
        self.tab2_site_proj_irrep1.clear()
        comb = [f"{i[0][0]}{no+1:02d}: {i[0]}" for no, i in enumerate(self.tab2_site_proj_comb_select)]
        self.tab2_site_proj_irrep1.addItems(comb)
        self.tab2_site_proj_irrep1.setCurrentIndex(0)

    # ==================================================
    def tab2_plot_site_cluster_samb(self):
        """
        plot site cluster SAMB.
        """
        if not hasattr(self, "tab2_site_proj_z_samb"):
            return
        irrep = self.tab2_site_proj_irrep1.currentIndex()

        v = NSArray.vector3d()
        cluster_obj = create_samb_object(
            self.tab2_site_proj_z_samb, self.tab2_site_proj_site, self.tab2_site_proj_c_samb, "Q", irrep, self._pgroup, v, False
        )

        label = self.tab2_site_proj_irrep1.currentText()[:3] + " \u21d0 " + remove_space(self.tab2_site_proj_pos.text())
        plot_site_cluster(self._qtdraw, self.tab2_site_proj_site, cluster_obj, label, self.pset)

    # ==================================================
    def tab2_gen_bond_cluster(self):
        """
        generate bond cluster SAMB.
        """
        pos = check_get_bond(self.tab2_bond_proj_pos.text())
        if pos is None:
            return

        combined_info = create_combined(self._group, self._pgroup, pos, 0, "Q", ret_bond=True)

        self.tab2_bond_proj_c_samb, self.tab2_bond_proj_site, self.tab2_bond_proj_z_samb = combined_info

        self.tab2_bond_cluster_samb_select()

    # ==================================================
    def tab2_bond_cluster_samb_select(self):
        """
        create bond cluster SAMB select list.
        """
        self.tab2_bond_proj_comb_select = self.tab2_bond_proj_z_samb["Q"] + self.tab2_bond_proj_z_samb["T"]
        qn = len(self.tab2_bond_proj_z_samb["Q"])
        self.tab2_bond_proj_irrep1.clear()
        comb = [
            f"Q{no+1:02d}: {i[0]}" if i[0][0] == "Q" else f"T{no+1-qn:02d}: {i[0]}"
            for no, i in enumerate(self.tab2_bond_proj_comb_select)
        ]
        self.tab2_bond_proj_irrep1.addItems(comb)
        self.tab2_bond_proj_irrep1.setCurrentIndex(0)

    # ==================================================
    def tab2_plot_bond_cluster_samb(self):
        """
        plot bond cluster SAMB.
        """
        if not hasattr(self, "tab2_bond_proj_z_samb"):
            return
        irrep = self.tab2_bond_proj_irrep1.currentIndex()
        z_head = self.tab2_bond_proj_irrep1.currentText()[0]
        if z_head == "T":
            irrep -= len(self.tab2_bond_proj_z_samb["Q"])
        t_odd = z_head != "Q"

        v = NSArray.vector3d()
        cluster_obj = create_samb_object(
            self.tab2_bond_proj_z_samb,
            self.tab2_bond_proj_site,
            self.tab2_bond_proj_c_samb,
            z_head,
            irrep,
            self._pgroup,
            v,
            t_odd,
        )

        label = self.tab2_bond_proj_irrep1.currentText()[:3] + " \u21d0 " + remove_space(self.tab2_bond_proj_pos.text())
        plot_bond_cluster(self._qtdraw, self.tab2_bond_proj_site, cluster_obj, label, self.pset, z_head)

    # ==================================================
    def tab2_gen_z_samb_vector(self):
        """
        generate vector cluster SAMB for given site/bond.
        """
        pos = check_get_site_bond(self.tab2_vector_proj_pos.text())
        if pos is None:
            return

        combined_info = create_combined(self._group, self._pgroup, pos, 1, self.tab2_vector_proj_type.currentText())

        self.tab2_vector_proj_c_samb, self.tab2_vector_proj_site, self.tab2_vector_proj_z_samb = combined_info

        self.tab2_select_z_samb_vector()

    # ==================================================
    def tab2_select_z_samb_vector(self):
        """
        create vector cluster SAMB select list.
        """
        self.tab2_vector_proj_comb_select = self.tab2_vector_proj_z_samb[self.tab2_vector_proj_type1.currentText()]
        self.tab2_vector_proj_irrep1.clear()
        comb = [f"{i[0][0]}{no+1:02d}: {i[0]}" for no, i in enumerate(self.tab2_vector_proj_comb_select)]
        self.tab2_vector_proj_irrep1.addItems(comb)
        self.tab2_vector_proj_irrep1.setCurrentIndex(0)

    # ==================================================
    def tab2_plot_z_samb_vector(self):
        """
        plot vector SAMB.
        """
        if not hasattr(self, "tab2_vector_proj_z_samb"):
            return
        z_head = self.tab2_vector_proj_type1.currentText()
        head = self.tab2_vector_proj_type.currentText()
        irrep = self.tab2_vector_proj_irrep1.currentIndex()
        t_odd = head.replace("M", "T").replace("G", "Q") != z_head.replace("M", "T").replace("G", "Q")

        v = NSArray.vector3d()
        cluster_obj = create_samb_object(
            self.tab2_vector_proj_z_samb,
            self.tab2_vector_proj_site,
            self.tab2_vector_proj_c_samb,
            z_head,
            irrep,
            self._pgroup,
            v,
            t_odd,
        )

        label = (
            self.tab2_vector_proj_irrep1.currentText()[:3]
            + " \u21d0 "
            + head
            + ", "
            + remove_space(self.tab2_vector_proj_pos.text())
        )
        plot_vector_cluster(self._qtdraw, self.tab2_vector_proj_site, cluster_obj, label, self.pset, head, v)

    # ==================================================
    def tab2_plot_z_samb_vector_lc(self):
        """
        plot linear combination of vector SAMB.
        """
        if not hasattr(self, "tab2_vector_proj_z_samb"):
            return
        head = self.tab2_vector_proj_type.currentText()
        form, ex_var, t_odd = check_linear_combination(self.tab2_vector_proj_z_samb, self.tab2_vector_proj_lc.text(), head)
        if form is None:
            return

        v = NSArray.vector3d()
        lc_basis = {
            i: sp.Matrix(
                create_samb_object(
                    self.tab2_vector_proj_z_samb,
                    self.tab2_vector_proj_site,
                    self.tab2_vector_proj_c_samb,
                    i[0].upper(),
                    int(i[1:]) - 1,
                    self._pgroup,
                    v,
                    t_odd,
                ).tolist()
            )
            for i in ex_var
        }
        cluster_obj = NSArray(str(NSArray(form).subs(lc_basis).tolist().T.tolist()[0]))

        label = (
            remove_space(self.tab2_vector_proj_lc.text())
            + " \u21d0 "
            + head
            + ", "
            + remove_space(self.tab2_vector_proj_pos.text())
        )

        plot_vector_cluster(self._qtdraw, self.tab2_vector_proj_site, cluster_obj, label, self.pset, head, v)

    # ==================================================
    def tab2_create_modulated_vector(self):
        """
        create modulated vector.
        """
        if not hasattr(self, "tab2_vector_proj_z_samb"):
            return
        if self.tab2_vector_modulation_type.currentIndex() == 0:
            basis = [f"Q{i+1:02d}" for i in range(len(self.tab2_vector_proj_z_samb["Q"]))] + [
                f"G{i+1:02d}" for i in range(len(self.tab2_vector_proj_z_samb["G"]))
            ]
        else:
            basis = [f"T{i+1:02d}" for i in range(len(self.tab2_vector_proj_z_samb["T"]))] + [
                f"M{i+1:02d}" for i in range(len(self.tab2_vector_proj_z_samb["M"]))
            ]
        if len(basis) < 1:
            return

        head = self.tab2_vector_proj_type.currentText()
        modulation = self.tab2_vector_modulation_list.text()

        if not self.tab2_modulated_vector_active:
            self.tab2_modulated_vector_active = True
            self.dialog_modulation_vector = DialogModulation(basis, modulation, head, False, self)
            self.dialog_modulation_vector.show()

    # ==================================================
    def tab2_gen_z_samb_orbital(self):
        """
        generate orbital cluster SAMB for given site/bond.
        """
        pos = check_get_site_bond(self.tab2_orbital_proj_pos.text())
        if pos is None:
            return

        combined_info = create_combined(
            self._group,
            self._pgroup,
            pos,
            self.tab2_orbital_proj_rank.currentText(),
            self.tab2_orbital_proj_type.currentText(),
        )

        self.tab2_orbital_proj_c_samb, self.tab2_orbital_proj_site, self.tab2_orbital_proj_z_samb = combined_info

        self.tab2_select_z_samb_orbital()

    # ==================================================
    def tab2_select_z_samb_orbital(self):
        """
        create orbital cluster SAMB select list.
        """
        self.tab2_orbital_proj_comb_select = self.tab2_orbital_proj_z_samb[self.tab2_orbital_proj_type1.currentText()]
        self.tab2_orbital_proj_irrep1.clear()
        comb = [f"{i[0][0]}{no+1:02d}: {i[0]}" for no, i in enumerate(self.tab2_orbital_proj_comb_select)]
        self.tab2_orbital_proj_irrep1.addItems(comb)
        self.tab2_orbital_proj_irrep1.setCurrentIndex(0)

    # ==================================================
    def tab2_plot_z_samb_orbital(self):
        """
        plot orbital SAMB.
        """
        if not hasattr(self, "tab2_orbital_proj_z_samb"):
            return
        z_head = self.tab2_orbital_proj_type1.currentText()
        head = self.tab2_orbital_proj_type.currentText()
        irrep = self.tab2_orbital_proj_irrep1.currentIndex()
        t_odd = head.replace("M", "T").replace("G", "Q") != z_head.replace("M", "T").replace("G", "Q")

        v = NSArray.vector3d()
        cluster_obj = create_samb_object(
            self.tab2_orbital_proj_z_samb,
            self.tab2_orbital_proj_site,
            self.tab2_orbital_proj_c_samb,
            z_head,
            irrep,
            self._pgroup,
            v,
            t_odd,
        )

        label = (
            self.tab2_orbital_proj_irrep1.currentText()[:3]
            + " \u21d0 "
            + head
            + self.tab2_orbital_proj_rank.currentText()
            + ", "
            + remove_space(self.tab2_orbital_proj_pos.text())
        )
        plot_orbital_cluster(self._qtdraw, self.tab2_orbital_proj_site, cluster_obj, label, self.pset, head)

    # ==================================================
    def tab2_plot_z_samb_orbital_lc(self):
        """
        plot linear combination of orbital SAMB.
        """
        if not hasattr(self, "tab2_orbital_proj_z_samb"):
            return
        head = self.tab2_orbital_proj_type.currentText()
        form, ex_var, t_odd = check_linear_combination(self.tab2_orbital_proj_z_samb, self.tab2_orbital_proj_lc.text(), head)
        if form is None:
            return

        v = NSArray.vector3d()
        lc_basis = {
            i: sp.Matrix(
                create_samb_object(
                    self.tab2_orbital_proj_z_samb,
                    self.tab2_orbital_proj_site,
                    self.tab2_orbital_proj_c_samb,
                    i[0].upper(),
                    int(i[1:]) - 1,
                    self._pgroup,
                    v,
                    t_odd,
                ).tolist()
            )
            for i in ex_var
        }
        cluster_obj = NSArray(str(NSArray(form).subs(lc_basis).tolist().T.tolist()[0]))

        label = (
            remove_space(self.tab2_orbital_proj_lc.text())
            + " \u21d0 "
            + head
            + self.tab2_orbital_proj_rank.currentText()
            + ", "
            + remove_space(self.tab2_orbital_proj_pos.text())
        )
        plot_orbital_cluster(self._qtdraw, self.tab2_orbital_proj_site, cluster_obj, label, self.pset, head)

    # ==================================================
    def tab2_create_modulated_orbital(self):
        """
        create modulated orbital.
        """
        if not hasattr(self, "tab2_orbital_proj_z_samb"):
            return
        if self.tab2_orbital_modulation_type.currentIndex() == 0:
            basis = [f"Q{i+1:02d}" for i in range(len(self.tab2_orbital_proj_z_samb["Q"]))] + [
                f"G{i+1:02d}" for i in range(len(self.tab2_orbital_proj_z_samb["G"]))
            ]
        else:
            basis = [f"T{i+1:02d}" for i in range(len(self.tab2_orbital_proj_z_samb["T"]))] + [
                f"M{i+1:02d}" for i in range(len(self.tab2_orbital_proj_z_samb["M"]))
            ]
        if len(basis) < 1:
            return

        head = self.tab2_orbital_proj_type.currentText()
        modulation = self.tab2_orbital_modulation_list.text()

        if not self.tab2_modulated_orbital_active:
            self.tab2_modulated_orbital_active = True
            self.dialog_modulation_orbital = DialogModulation(basis, modulation, head, True, self)
            self.dialog_modulation_orbital.show()

    # ==================================================
    def tab2_plot_z_samb_hopping(self):
        """
        plot hopping direction.
        """
        pos = check_get_bond(self.tab2_hopping_proj_pos.text())
        if pos is None:
            return

        head = "T"
        combined_info = create_combined(self._group, self._pgroup, pos, 1, head)

        c_samb, site, z_samb = combined_info

        v = NSArray.vector3d()
        cluster_obj = create_samb_object(z_samb, site, c_samb, "Q", 0, self._pgroup, v, True)

        label = "t_dir \u21d0 " + remove_space(self.tab2_hopping_proj_pos.text())
        plot_vector_cluster(self._qtdraw, site, cluster_obj, label, self.pset, head, v)

    # ==================================================
    def save_dict(self):
        """
        create qtdraw dict (MultiPie part).

        Returns:
            dict: qtdraw-multipie dict.
        """
        dic = {
            "version": self._qtdraw._multipie_loaded,
            "main": {"group": (self.main_g_type.currentIndex(), self.main_c_type.currentIndex(), self.main_group.currentIndex())},
            "tab1": {
                "site": (self.tab1_site_pos.text(),),
                "bond": (self.tab1_bond_pos.text(),),
                "vector": (self.tab1_vector_pos.text(), self.tab1_vector_type.currentIndex()),
                "orbital": (self.tab1_orbital_pos.text(), self.tab1_orbital_type.currentIndex()),
                "harmonics": (
                    self.tab1_pgharm_pos.text(),
                    self.tab1_pgharm_type.currentIndex(),
                    self.tab1_pgharm_rank.currentIndex(),
                    self.tab1_pgharm_irrep.currentIndex(),
                ),
            },
            "tab2": {
                "site": (self.tab2_site_proj_pos.text(), self.tab2_site_proj_irrep1.currentIndex()),
                "bond": (self.tab2_bond_proj_pos.text(), self.tab2_bond_proj_irrep1.currentIndex()),
                "vector": (
                    self.tab2_vector_proj_pos.text(),
                    self.tab2_vector_proj_type.currentIndex(),
                    self.tab2_vector_proj_type1.currentIndex(),
                    self.tab2_vector_proj_irrep1.currentIndex(),
                    self.tab2_vector_proj_lc.text(),
                    self.tab2_vector_modulation_type.currentIndex(),
                    self.tab2_vector_modulation_list.text(),
                ),
                "orbital": (
                    self.tab2_orbital_proj_pos.text(),
                    self.tab2_orbital_proj_type.currentIndex(),
                    self.tab2_orbital_proj_rank.currentIndex(),
                    self.tab2_orbital_proj_type1.currentIndex(),
                    self.tab2_orbital_proj_irrep1.currentIndex(),
                    self.tab2_orbital_proj_lc.text(),
                    self.tab2_orbital_modulation_type.currentIndex(),
                    self.tab2_orbital_modulation_list.text(),
                ),
                "hopping": (self.tab2_hopping_proj_pos.text(),),
            },
        }

        return dic

    # ==================================================
    def load_dict(self, dic):
        """
        load dict and set.

        Args:
            dic (dict): loaded qtdraw dict.
        """
        group_type, crystal_type, group = dic["main"]["group"]
        self.main_g_type.setCurrentIndex(group_type)
        self.main_c_type.setCurrentIndex(crystal_type)
        self.main_group.setCurrentIndex(group)

        t1_site_pos = dic["tab1"]["site"][0]
        self.tab1_site_pos.setText(t1_site_pos)
        t1_bond_pos = dic["tab1"]["bond"][0]
        self.tab1_bond_pos.setText(t1_bond_pos)
        t1_vector_pos, t1_vector_type = dic["tab1"]["vector"]
        self.tab1_vector_pos.setText(t1_vector_pos)
        self.tab1_vector_type.setCurrentIndex(t1_vector_type)
        t1_orbital_pos, t1_orbital_type = dic["tab1"]["orbital"]
        self.tab1_orbital_pos.setText(t1_orbital_pos)
        self.tab1_orbital_type.setCurrentIndex(t1_orbital_type)
        t1_harmonics_pos, t1_harmonics_type, t1_harmonics_rank, t1_harmonics_irrep = dic["tab1"]["harmonics"]
        self.tab1_pgharm_pos.setText(t1_harmonics_pos)
        self.tab1_pgharm_type.setCurrentIndex(t1_harmonics_type)
        self.tab1_pgharm_rank.setCurrentIndex(t1_harmonics_rank)
        self.tab1_pgharm_irrep.setCurrentIndex(t1_harmonics_irrep)

        t2_site_pos, t2_site_irrep1 = dic["tab2"]["site"]
        self.tab2_site_proj_pos.setText(t2_site_pos)
        self.tab2_site_proj_pos.returnPressed.emit()
        self.tab2_site_proj_irrep1.setCurrentIndex(t2_site_irrep1)
        t2_bond_pos, t2_bond_irrep1 = dic["tab2"]["bond"]
        self.tab2_bond_proj_pos.setText(t2_bond_pos)
        self.tab2_bond_proj_pos.returnPressed.emit()
        self.tab2_bond_proj_irrep1.setCurrentIndex(t2_bond_irrep1)
        (
            t2_vector_pos,
            t2_vector_type,
            t2_vector_type1,
            t2_vector_irrep1,
            t2_vector_lc,
            t2_vector_mod_type,
            t2_vector_mod_list,
        ) = dic["tab2"]["vector"]
        self.tab2_vector_proj_pos.setText(t2_vector_pos)
        self.tab2_vector_proj_type.setCurrentIndex(t2_vector_type)
        self.tab2_vector_proj_pos.returnPressed.emit()
        self.tab2_vector_proj_type1.setCurrentIndex(t2_vector_type1)
        self.tab2_vector_proj_irrep1.setCurrentIndex(t2_vector_irrep1)
        self.tab2_vector_proj_lc.setText(t2_vector_lc)
        self.tab2_vector_modulation_type.setCurrentIndex(t2_vector_mod_type)
        self.tab2_vector_modulation_list.setText(t2_vector_mod_list)
        (
            t2_orbital_pos,
            t2_orbital_type,
            t2_orbital_rank,
            t2_orbital_type1,
            t2_orbital_irrep1,
            t2_orbital_lc,
            t2_orbital_mod_type,
            t2_orbital_mod_list,
        ) = dic["tab2"]["orbital"]
        self.tab2_orbital_proj_pos.setText(t2_orbital_pos)
        self.tab2_orbital_proj_type.setCurrentIndex(t2_orbital_type)
        self.tab2_orbital_proj_rank.setCurrentIndex(t2_orbital_rank)
        self.tab2_orbital_proj_pos.returnPressed.emit()
        self.tab2_orbital_proj_type1.setCurrentIndex(t2_orbital_type1)
        self.tab2_orbital_proj_irrep1.setCurrentIndex(t2_orbital_irrep1)
        self.tab2_orbital_proj_lc.setText(t2_orbital_lc)
        self.tab2_orbital_modulation_type.setCurrentIndex(t2_orbital_mod_type)
        self.tab2_orbital_modulation_list.setText(t2_orbital_mod_list)
        t2_hopping_pos = dic["tab2"]["hopping"][0]
        self.tab2_hopping_proj_pos.setText(t2_hopping_pos)



from qtpy.QtWidgets import QColorDialog
from qtdraw.core.color_palette import hex_colornames, _hex2rgb


# ==================================================
class QtColorDialog(QColorDialog):
    """
    color dialog.

    Attributes:
        hex (str): hex code of color
        name (str): color name
        rgb (tuple): RGB color
    """

    # ==================================================
    def __init__(self, parent=None):
        """
        initialize the class.
        """
        super().__init__(parent)
        self.finished["int"].connect(self._close)

    # ==================================================
    def _close(self):
        color = self.selectedColor()

        self.hex = color.name()
        self.rgb = _hex2rgb(self.hex)
        if self.hex in hex_colornames.keys():
            self.name = hex_colornames[self.hex]
        else:
            self.name = ""

        print(self.name, self.hex, self.rgb)



import numpy as np
import sympy as sp
from gcoreutils.nsarray import NSArray
from gcoreutils.convert_util import text_to_list
from gcoreutils.string_util import remove_space
from multipie.model.material_model import MaterialModel
from qtdraw.multipie.setting import rcParams


# ==================================================
def check_get_site(txt):
    """
    check and get site.

    Args:
        txt (str): site.

    Returns:
        NSArray: site or None if invalid.
    """
    try:
        site = NSArray(txt)
        if site.style != "vector":
            return None
    except Exception:
        return None

    return site


# ==================================================
def check_get_bond(txt):
    """
    check and get bond.

    Args:
        txt (str): bond.

    Returns:
        NSArray: bond or None if invalid.
    """
    try:
        bond = NSArray(txt)
        if bond.style not in ["bond", "bond_th", "bond_sv"]:
            return None
    except Exception:
        return None

    return bond


# ==================================================
def check_get_site_bond(txt):
    """
    check and get site (bond center) from site_bond.

    Args:
        txt (str): site_bond.

    Returns:
        NSArray: site (bond center), or None if invalid.
    """
    try:
        site_bond = NSArray(txt)
        if site_bond.style not in ["vector", "bond", "bond_th", "bond_sv"]:
            return None
    except Exception:
        return None

    return site_bond


# ==================================================
def create_combined(group, pg, site_bond, rank, head, ret_bond=False):
    """
    create SAMB.

    Args:
        group (Group): space/point group.
        pg (PointGroup): associated point group.
        site_bond (str): site or bond.
        rank (str): harmonics rank.
        head (str): harmonics type.
        ret_bond (bool, optional): return bond ?

    Returns:
        tuple:
            - dict: cluster SAMB.
            - NSArray: site or bond.
            - dict: combined SAMB.
    """
    t_rev = {"Q": "Q", "G": "G", "T": "Q", "M": "G"}

    if site_bond.style == "vector":
        c_samb, site = group.site_cluster_samb(site_bond)
    else:
        c_samb, bond = group.bond_cluster_samb(site_bond)
        if ret_bond:
            site = bond
        else:
            site = bond.convert_bond("bond")[1]

    x_tag = pg.harmonics.key_list().select(rank=int(rank), head=t_rev[head])
    if head in ["T", "M"]:
        x_tag = [tag.reverse_t_type() for tag in x_tag]
    y_tag = list(c_samb.keys())

    z_samb_all = group.z_samb(x_tag, y_tag)
    z_samb = {"Q": [], "G": [], "T": [], "M": []}
    for tag, c in z_samb_all.items():
        tag_str = combined_format(tag)
        z_samb[tag[0].head].append((tag_str, c))
    for k in z_samb.keys():
        z_samb[k] = list(sorted(z_samb[k], key=lambda i: i[0]))

    return c_samb, site, z_samb


# ==================================================
def create_samb_object(z_samb, site, c_samb, z_head, irrep, pg, v, t_odd):
    """
    create SAMB object.

    Args:
        z_samb (dict): combined SAMB.
        site (NSArray): site.
        c_samb (dict): cluster SAMB.
        z_head (str): multipole type.
        irrep (int): irrep. index.
        pg (PointGroup): point group.
        v (NSArray): vector variable.
        t_odd (bool): magnetic bond ?

    Returns:
        NSArray: (xyz)-polynomial at each cluster site.
    """
    obj = NSArray.zeros(len(site), "vector")
    for i in z_samb[z_head][irrep][1]:
        coeff, tag_h, tag_c = i
        harm = pg.harmonics[tag_h].expression(v=v)
        cluster = c_samb[tag_c]
        obj += coeff * harm * cluster

    if t_odd:
        obj *= -sp.I

    return obj


# ==================================================
def check_linear_combination(z_samb, form, head):
    """
    check form of linear combination.

    Args:
        z_samb (dict): combined SAMB.
        form (str): expression of linear combination.
        head (str): multipole type.

    Returns:
        tuple:
            - str: expression of linear combination (lower case).
            - set: used variables.
            - bool: magnetic bond ?
    """
    irrep_num = {i: len(z_samb[i]) for i in ["Q", "G", "T", "M"]}
    var_e = set([f"q{i+1:02d}" for i in range(irrep_num["Q"])] + [f"g{i+1:02d}" for i in range(irrep_num["G"])])
    var_m = set([f"t{i+1:02d}" for i in range(irrep_num["T"])] + [f"m{i+1:02d}" for i in range(irrep_num["M"])])

    form = form.lower()
    form_variable = set(NSArray(form).variable())

    t_odd = "Q"
    if form_variable.issubset(var_m):
        t_odd = "T"
    elif not form_variable.issubset(var_e):
        return None, None, None
    t_odd = head.replace("M", "T").replace("G", "Q") != t_odd

    return form, form_variable, t_odd


# ==================================================
def combined_format(tag_list):
    """
    create formatted combined SAMB.

    Args:
        tag_list (tuple): (Z,X,Y) tag.

    Returns:
        str: formatted combined SAMB.
    """
    z_tag, x_tag, y_tag = tag_list
    t1 = (",".join(str(x_tag).split(",")[:-1]) + ")").replace("h", "a")
    t2 = ",".join(str(y_tag).split(",")[:-1]) + ")"
    tag = f"{z_tag} = {t1} x {t2}"

    return tag


# ==================================================
def plot_equivalent_site(qtdraw, site, n_pset):
    """
    plot equivalent sites.

    Args:
        qtdraw (QtDraw): QtDraw.
        site (NSArray): equivalent sites (including plus_set).
        n_pset (int): no. of plus set.
    """
    name0 = qtdraw._get_name("site")
    name0 = f"S{int(name0[1:])+1}"
    qtdraw._close_dialog()
    color = rcParams["site_color"]
    primitive_num = len(site) // n_pset

    for no, (s, mp) in enumerate(site.items()):
        mp = MaterialModel._mapping_str(mp)
        idx = no % primitive_num
        pset = no // primitive_num
        name = name0
        if n_pset > 1:
            name += f"({pset+1})"
        label = f"s{idx+1}:{mp}"
        qtdraw.plot_site(s, color=color, name=name, label=label, show_lbl=rcParams["show_label"])
    qtdraw._plot_all_object()


# ==================================================
def plot_equivalent_bond(qtdraw, bond, nondirectional, n_pset):
    """
    plot equivalent bonds.

    Args:
        qtdraw (QtDraw): QtDraw.
        bond (NSArray): equivalent bonds (including plus_set).
        nondirectional (bool): nondirectional ?
        n_pset (int): no. of plus set.
    """
    name0 = qtdraw._get_name("bond")
    name0 = f"B{int(name0[1:])+1}"
    qtdraw._close_dialog()
    color1 = rcParams["bond_color1"]
    if nondirectional:
        color2 = color1
    else:
        color2 = rcParams["bond_color2"]
    primitive_num = len(bond) // n_pset

    for no, (b, mp) in enumerate(bond.items()):
        b = NSArray(b)
        v, c = b.convert_bond("bond")
        mp = MaterialModel._mapping_str(mp)
        idx = no % primitive_num
        pset = no // primitive_num
        name = name0
        if n_pset > 1:
            name += f"({pset+1})"
        label = f"b{idx+1}:{mp}"
        qtdraw.plot_bond(c, v, color=color1, color2=color2, name=name, label=label, show_lbl=rcParams["show_label"])
    qtdraw._plot_all_object()


# ==================================================
def plot_vector_equivalent_site(qtdraw, site, vector, head, n_pset):
    """
    plot vectors at equivalent sites.

    Args:
        qtdraw (QtDraw): QtDraw.
        site (NSArray): equivalent sites (including plus_set).
        vector (NSArray): vector.
        head (str): multipole type.
        n_pset (int): no. of plus set.
    """
    name0 = qtdraw._get_name("vector")
    name0 = f"V{int(name0[3:])+1}"
    qtdraw._close_dialog()
    color = rcParams["vector_color_" + head]
    primitive_num = len(site) // n_pset

    for no, s in enumerate(site):
        idx = no % primitive_num
        pset = no // primitive_num
        name = name0
        if n_pset > 1:
            name += f"({pset+1})"
        label = f"v{idx+1}"
        qtdraw.plot_vector(s, vector, color=color, name=name, label=label, show_lbl=rcParams["show_label"])
    qtdraw._plot_all_object()


# ==================================================
def plot_orbital_equivalent_site(qtdraw, site, orbital, head, n_pset):
    """
    plot orbitals at equivalent sites.

    Args:
        qtdraw (QtDraw): QtDraw.
        site (NSArray): equivalent sites (including plus_set).
        orbital (str): orbital.
        head (str): multipole type.
        n_pset (int): no. of plus set.
    """
    name0 = qtdraw._get_name("orbital")
    name0 = f"O{int(name0[3:])+1}"
    qtdraw._close_dialog()
    color = rcParams["orbital_color_" + head]
    primitive_num = len(site) // n_pset

    for no, s in enumerate(site):
        idx = no % primitive_num
        pset = no // primitive_num
        name = name0
        if n_pset > 1:
            name += f"({pset+1})"
        label = f"o{idx+1}"
        qtdraw.plot_orbital(s, orbital, size=0.3, color=color, name=name, label=label, show_lbl=rcParams["show_label"])
    qtdraw._plot_all_object()


# ==================================================
def plot_site_cluster(qtdraw, site, obj, label, pset):
    """
    plot site cluster SAMB.

    Args:
        qtdraw (QtDraw): QtDraw.
        site (NSArray): equivalent sites.
        obj (NSArray): SAMB weight.
        label (str): label.
        pset (NSArray): plus set.
    """
    color = []
    for w in obj:
        if w > 0:
            c = "salmon"
        elif w < 0:
            c = "aqua"
        else:
            c = "silver"
        color.append(c)

    obj /= np.abs(obj).max()
    n_pset = len(pset)

    qtdraw._close_dialog()
    name = "Z_" + qtdraw._get_name("site")
    for no, p in enumerate(pset):
        name1 = name
        if n_pset != 1:
            name1 += f"({no+1})"
        for s, w, cl in zip(site, obj, color):
            s = (s + p).shift()
            if cl == "silver":
                w = 1
            qtdraw.plot_site(
                s,
                size=abs(w),
                color=cl,
                name=name1,
                label=label,
                show_lbl=rcParams["show_label"],
            )
    qtdraw._plot_all_object()


# ==================================================
def plot_bond_cluster(qtdraw, bond, obj, label, pset, z_head):
    """
    plot bond cluster SAMB.

    Args:
        qtdraw (QtDraw): QtDraw.
        bond (NSArray): equivalent bonds.
        obj (NSArray): SAMB weight.
        label (str): label.
        pset (NSArray): plus set.
        z_head (str): multipole type.
    """
    color = []
    if z_head == "Q":
        for w in obj:
            if w > 0:
                c = "salmon"
            elif w < 0:
                c = "aqua"
            else:
                c = "silver"
            color.append(c)
    else:
        for w in obj:
            if w == 0:
                c = "silver"
            else:
                c = "salmon"
            color.append(c)

    obj /= np.abs(obj).max()
    n_pset = len(pset)

    qtdraw._close_dialog()
    name = "Z_" + qtdraw._get_name("bond")
    if z_head == "Q":
        for no, p in enumerate(pset):
            name1 = name
            if n_pset != 1:
                name1 += f"({no+1})"
            for s, w, cl in zip(bond, obj, color):
                v, c = s.convert_bond("bond")
                c = (c + p).shift()
                if cl == "silver":
                    w = 1
                qtdraw.plot_bond(
                    c,
                    v,
                    color=cl,
                    color2=cl,
                    width=abs(w),
                    name=name1,
                    label=label,
                    show_lbl=rcParams["show_label"],
                )
    else:
        for no, p in enumerate(pset):
            name1 = name
            if n_pset != 1:
                name1 += f"({no+1})"
            for s, w, cl in zip(bond, obj, color):
                v, c = s.convert_bond("bond")
                c = (c + p).shift()
                if cl == "silver":
                    w = 1
                    qtdraw.plot_bond(
                        c,
                        v,
                        color=cl,
                        color2=cl,
                        width=abs(w),
                        name=name1,
                        label=label,
                        show_lbl=rcParams["show_label"],
                    )
                else:
                    v = v.transform(qtdraw._A)
                    if w < 0:
                        v = -v
                    norm = v.norm() * 0.7
                    qtdraw.plot_vector(
                        c,
                        v,
                        color=cl,
                        width=abs(w),
                        length=norm,
                        offset=-0.5,
                        name=name1,
                        label=label,
                        show_lbl=rcParams["show_label"],
                    )

    qtdraw._plot_all_object()


# ==================================================
def plot_vector_cluster(qtdraw, site, obj, label, pset, head, v):
    """
    plot vector cluster SAMB.

    Args:
        qtdraw (QtDraw): QtDraw.
        site (NSArray): equivalent sites.
        obj (NSArray): SAMB weight.
        label (str): label.
        pset (NSArray): plus set.
        head (str): multipole type.
        v (NSArray): vector variable.
    """
    rep = {v[0]: sp.Matrix([1, 0, 0]), v[1]: sp.Matrix([0, 1, 0]), v[2]: sp.Matrix([0, 0, 1])}
    color = rcParams["vector_color_" + head]
    qtdraw._close_dialog()
    name = "Z_" + qtdraw._get_name("vector")
    n_pset = len(pset)

    for no, p in enumerate(pset):
        name1 = name
        if n_pset != 1:
            name1 += f"({no+1})"
        for s, c in zip(site, obj):
            s = (s + p).shift()
            if c != 0:
                c = str(c.subs(rep).T[:])
                c = NSArray(c)
                d = c.norm()
                qtdraw.plot_vector(s, c, length=d, color=color, name=name1, label=label, show_lbl=rcParams["show_label"])

    qtdraw._plot_all_object()


# ==================================================
def plot_orbital_cluster(qtdraw, site, obj, label, pset, head):
    """
    plot orbital cluster SAMB.

    Args:
        qtdraw (QtDraw): QtDraw.
        site (NSArray): equivalent sites.
        obj (NSArray): SAMB weight.
        label (str): label.
        pset (NSArray): plus set.
        head (str): multipole type.
    """
    color = rcParams["orbital_color_" + head]
    qtdraw._close_dialog()
    name = "Z_" + qtdraw._get_name("orbital")
    n_pset = len(pset)

    for no, p in enumerate(pset):
        name1 = name
        if n_pset != 1:
            name1 += f"({no+1})"
        for s, orb in zip(site, obj):
            s = (s + p).shift()
            qtdraw.plot_orbital(
                s,
                orb,
                size=0.6,
                scale=False,
                color=color,
                name=name1,
                label=label,
                show_lbl=rcParams["show_label"],
            )

    qtdraw._plot_all_object()


# ==================================================
def create_phase_factor(modulation, repeat, offset, pset):
    """
    create phase factor.

    Args:
        modulation (list): modulation info. [(basis, coeff, k, n)]
        repeat (list): repeat.
        offset (list): offset.
        pset (NSArray): plus set.

    Returns: tuple.
        - dict: {(k(str),n(int),plus_set no(int)): [phase at each grid(float)]}.
        - list: cell grid.
    """
    igrid = NSArray.igrid(repeat, offset)
    pset = np.array(pset.tolist(), dtype=float)
    phase_dict = {}
    for _, _, k, n in modulation:
        kvec = np.array(NSArray(k).tolist(), dtype=float)
        for p_no, p in enumerate(pset):
            lst = []
            for i in igrid:
                kr = 2.0 * np.pi * kvec @ (i.numpy().astype(float) + p)
                phase = np.cos(kr) if n == 0 else np.sin(kr)
                lst.append(phase)
            phase_dict[(k, n, p_no)] = lst

    return phase_dict, igrid.numpy().astype(int).tolist()


# ==================================================
def create_modulated_samb_object(z_samb, site, c_samb, pg, v, head, modulation, repeat, offset, pset):
    """
    create modulated SAMB object.

    Args:
        z_samb (dict): combined SAMB.
        site (NSArray): site.
        c_samb (dict): cluster SAMB.
        pg (PointGroup): point group.
        v (NSArray): vector variable.
        head (str): _description_
        modulation (list): modulation info.
        repeat (list): repeat.
        offset (list): offset.
        pset (NSArray): plus set.

    Returns: tuple.
        - list: [(xyz)-polynomial at each cluster site.] [cell][pset].
        - list: cell grid.
    """
    phase_dict, igrid = create_phase_factor(modulation, repeat, offset, pset)
    obj = NSArray.zeros((len(igrid) * len(pset), len(site)), "vector")
    for p_no in range(len(pset)):
        for basis, coeff, k, n in modulation:
            z_head = basis[0]
            irrep = int(basis[1:]) - 1
            t_odd = head.replace("M", "T").replace("G", "Q") != z_head.replace("M", "T").replace("G", "Q")
            phase = phase_dict[(k, n, p_no)]
            coeff = NSArray(coeff, fmt="value")
            cluster_obj = create_samb_object(
                z_samb,
                site,
                c_samb,
                z_head,
                irrep,
                pg,
                v,
                t_odd,
            )
            for i_no in range(len(igrid)):
                obj[i_no * len(pset) + p_no] += coeff * phase[i_no] * cluster_obj

    obj = obj.numpy().reshape((len(igrid), len(pset), len(site))).tolist()

    return obj, igrid


# ==================================================
def plot_modulated_vector_cluster(qtdraw, site, obj, name, pset, igrid, head, v):
    """
    plot modulated vector cluster SAMB.

    Args:
        qtdraw (QtDraw): QtDraw.
        site (NSArray): equivalent sites.
        obj (list): SAMB weight at (cell,pset).
        name (str): name.
        pset (NSArray): plus set.
        igrid (NSArray): cell grid.
        head (str): multipole type.
        v (NSArray): vector variable.
    """
    rep = {v[0]: sp.Matrix([1, 0, 0]), v[1]: sp.Matrix([0, 1, 0]), v[2]: sp.Matrix([0, 0, 1])}
    color = rcParams["vector_color_" + head]
    qtdraw._close_dialog()
    n_pset = len(pset)

    for i_no, i in enumerate(igrid):
        for p_no, p in enumerate(pset):
            label = ""
            if n_pset != 1:
                label = f"({p_no+1})"
            for no, (s, c) in enumerate(zip(site, obj[i_no][p_no])):
                s = (s + p).shift()
                label = f"s{no+1}" + label
                if c != 0:
                    c = str(c.subs(rep).T[:])
                    c = NSArray(c)
                    d = c.norm()
                    qtdraw.plot_vector(
                        s, c, length=d, color=color, name=name, label=label, show_lbl=rcParams["show_label"], cell=i
                    )

    qtdraw._plot_all_object()


# ==================================================
def plot_modulated_orbital_cluster(qtdraw, site, obj, name, pset, igrid, head):
    """
    plot modulated vector cluster SAMB.

    Args:
        qtdraw (QtDraw): QtDraw.
        site (NSArray): equivalent sites.
        obj (list): SAMB weight at (cell,pset).
        name (str): name.
        pset (NSArray): plus set.
        igrid (NSArray): cell grid.
        head (str): multipole type.
    """
    color = rcParams["orbital_color_" + head]
    qtdraw._close_dialog()
    n_pset = len(pset)

    for i_no, i in enumerate(igrid):
        for p_no, p in enumerate(pset):
            label = ""
            if n_pset != 1:
                label += f"({p_no+1})"
            for no, (s, orb) in enumerate(zip(site, obj[i_no][p_no])):
                s = (s + p).shift()
                label = f"s{no+1}" + label
                qtdraw.plot_orbital(
                    s, orb, size=0.6, scale=False, color=color, name=name, label=label, show_lbl=rcParams["show_label"], cell=i
                )

    qtdraw._plot_all_object()


# ==================================================
def parse_modulation_list(lst):
    """
    parse modulation list.

    Args:
        lst (str): modulation list in str, [[basis,coeff,k,cos/sin]]

    Returns:
        list: modulation list.
    """
    lst = text_to_list(remove_space(lst))
    if lst is None:
        return None

    modulation = []
    for basis, coeff, k, cs in lst:
        cs = 0 if cs == "cos" else 1
        modulation.append([coeff, basis, "[" + (",".join(k)) + "]", cs])

    return modulation



"""
This file manages setting of QtDraw with MultiPie and its viewer style.
"""


# ==================================================
rcParams = {
    "group_tag": "C1",
    "show_label": False,
    "site_color": "darkseagreen",
    "bond_color1": "silver",
    "bond_color2": "iron",
    "vector_color_Q": "orange",
    "vector_color_M": "lightskyblue",
    "vector_color_T": "hotpink",
    "vector_color_G": "yellowgreen",
    "orbital_color_Q": "Wistia",
    "orbital_color_M": "GnBu",
    "orbital_color_T": "coolwarm",
    "orbital_color_G": "PiYG",
}


# ==================================================
default_style = {
    "site": [  # (color, size, opacity)
        ("darkseagreen", 1.0, 1.0),  # 1st site
        ("lightblue", 1.0, 1.0),  # 2nd site
        ("sandybrown", 1.0, 1.0),  # 9th site
        ("gold", 1.0, 1.0),  # 3rd site
        ("darkkhaki", 1.0, 1.0),  # 4th site
        ("skyblue", 1.0, 1.0),  # 7th site
        ("thistle", 1.0, 1.0),  # 6th site
        ("darkgrey", 1.0, 1.0),  # 8th site
        ("burlywood", 1.0, 1.0),  # 5th site
        ("ghostwhite", 1.0, 1.0),  # other sites
    ],
    "bond": [  # ((tail-color, head-color), width, opacity)
        (("snow", "silver"), 1.0, 1.0),  # 1st neighbor
        (("lightcyan", "lightsteelblue"), 1.0, 1.0),  # 2nd neighbor
        (("antiquewhite", "burlywood"), 1.0, 1.0),  # 3rd neighbor
        (("palegoldenrod", "darkseagreen"), 1.0, 1.0),  # 4th neighbor
        (("mistyrose", "lightpink"), 1.0, 1.0),  # 5th neighbor
        (("aliceblue", "lightblue"), 1.0, 1.0),  # 6th neighbor
        (("wheat", "sandybrown"), 1.0, 1.0),  # 7th neighbor
        (("seashell", "thistle"), 1.0, 1.0),  # 8th neighbor
        (("cornsilk", "peachpuff"), 1.0, 1.0),  # 9th neighbor
        (("whitesmoke", "darkkhaki"), 1.0, 1.0),  # other neighbors
    ],
}



from qtpy.QtCore import Qt, QSize, QModelIndex
from qtpy.QtWidgets import QItemDelegate, QStyle
from qtpy.QtGui import QColor
from qtdraw.core.tree_item_model import TreeItemModel
from qtdraw.core.editable_widget import (
    EditableWidget,
    QtText,
    QtImage,
    QtMath,
    QtCheckBox,
    QtRadioGroup,
    QtComboBox,
    QtColorSelector,
)
from qtdraw.core.setting import rcParams
from qtdraw.core.color_palette import all_colors


# ==================================================
class GroupModel(TreeItemModel):
    # ==================================================
    def __init__(self, ds, name, read_only=False, align=None, parent=None):
        roles = ds.role[name]
        if roles is None:
            roles = [""] * ds[name].shape[1]
        if align is None:
            align = ["left"] * ds[name].shape[1]

        self.parent_widget = parent
        self.read_only = read_only
        self.roles = []
        self.roles_opt = []
        self.init = []
        self.hide = []
        for i, role in enumerate(roles):
            self.roles.append(role[0])
            self.roles_opt.append(role[1:-1])
            self.init.append(role[-1])
            if role[0] == "hide":
                self.hide.append(i)
        self.align = align

        super().__init__(ds, name, self.init, parent)

    # ==================================================
    def _associated_index(self, index, key):
        if key not in self.header:
            return QModelIndex()
        else:
            return index.siblingAtColumn(self.header.index(key))

    # ==================================================
    def setDisplayDataRaw(self, index):
        # return QtText(str((index.row(), index.column())) + " => " + str(self.absIndex(index)), parent=self.parent_widget)
        # return QtText(str(index.data(Qt.UserRole)), parent=self.parent_widget)

        def _check(index_s, flag, index, data):
            self.setData(index_s, flag, emit=False)
            self.setData(index, data)

        parent = self.parent_widget
        data = index.data(Qt.UserRole)
        role = self.roles[index.column()]
        opt = self.roles_opt[index.column()]
        align = self.align[index.column()]

        if role == "check":
            if len(opt) == 0:
                item = QtText(
                    data, callback=lambda txt: self.setData(index, txt), read_only=self.read_only, align=align, parent=parent
                )
            else:
                index_s = self._associated_index(index, opt[0])  # check status.
                if index_s.column() == index.column():  # check only.
                    item = QtCheckBox(
                        current=data,
                        callback_check=lambda flag: self.setData(index_s, flag),
                        read_only=self.read_only,
                        parent=parent,
                    )
                else:  # check with text.
                    status = index_s.data(Qt.UserRole)
                    item = QtCheckBox(
                        label=data,
                        current=status,
                        callback_check=lambda flag: _check(index_s, flag, index, data),
                        read_only=self.read_only,
                        parent=parent,
                    )
        elif role == "check_edit":
            index_s = self._associated_index(index, opt[0])  # check status.
            status = index_s.data(Qt.UserRole)
            item = QtCheckBox(
                label=data,
                current=status,
                callback_check=lambda flag: _check(index_s, flag, index, data),
                callback_label=lambda text: self.setData(index, text, emit=True),
                read_only=self.read_only,
                parent=parent,
            )
        elif role == "text":
            if len(opt) == 0:  # no validator.
                data = str(data).replace("j", "i").strip(" ()")
                item = QtText(
                    data, callback=lambda txt: self.setData(index, txt), read_only=self.read_only, align=align, parent=parent
                )
            elif len(opt) == 1:  # validator only.
                item = QtText(
                    data,
                    callback=lambda txt: self.setData(index, txt),
                    validator=opt[0],
                    read_only=self.read_only,
                    align=align,
                    parent=parent,
                )
            else:  # validator + decimal
                item = QtText(
                    data,
                    callback=lambda txt: self.setData(index, txt),
                    validator=opt[0],
                    decimal=opt[1],
                    read_only=self.read_only,
                    align=align,
                    parent=parent,
                )
        elif role == "image":
            w = rcParams["detail.image.width"]
            item = QtImage(data, width=w, read_only=self.read_only, parent=parent)
        elif role == "math":
            color = rcParams["detail.latex.color"]
            style = rcParams["detail.latex.format"]
            dpi = rcParams["detail.latex.dpi"]
            if len(opt) == 0:  # no validator.
                item = QtMath(
                    data,
                    callback=lambda txt: self.setData(index, txt),
                    color=color,
                    style=style,
                    dpi=dpi,
                    read_only=self.read_only,
                    align=align,
                    parent=parent,
                )
            elif len(opt) == 1:  # validator only.
                item = QtMath(
                    data,
                    callback=lambda txt: self.setData(index, txt),
                    validator=opt[0],
                    color=color,
                    style=style,
                    dpi=dpi,
                    read_only=self.read_only,
                    align=align,
                    parent=parent,
                )
            else:  # validator + acceptable variables.
                item = QtMath(
                    data,
                    callback=lambda txt: self.setData(index, txt),
                    validator=opt[0],
                    check_var=opt[1],
                    color=color,
                    style=style,
                    dpi=dpi,
                    read_only=self.read_only,
                    align=align,
                    parent=parent,
                )
        elif role in ["color", "colormap", "color_both"]:
            item = QtColorSelector(
                current=data,
                color_type=role,
                callback=lambda color: self.setData(index, color),
                parent=parent,
            )
        elif role == "radio":
            lst = opt[0]
            vertical = opt[1] == "V"
            item = QtRadioGroup(
                lst=lst,
                current=data,
                vertical=vertical,
                callback=lambda idx: self.setData(index, idx),
                parent=parent,
            )
        elif role == "combo":
            lst = opt[0]
            item = QtComboBox(lst=lst, current=data, callback=lambda idx: self.setData(index, idx), parent=parent)
        elif role == "hide":
            return QtText("", read_only=True, parent=parent)
        else:
            item = QtText(
                str(data), callback=lambda txt: self.setData(index, txt), read_only=self.read_only, align=align, parent=parent
            )

        if not isinstance(item, EditableWidget):
            item = QtText("unknown widget", read_only=True, align=align, parent=parent)

        return item


# ==================================================
class Delegate(QItemDelegate):
    def __init__(self, bgcolor, parent=None):
        super().__init__(parent)
        self.bgcolor = QColor(all_colors[bgcolor][0])

    # ==================================================
    def paint(self, painter, option, index):

        if option.state & QStyle.State_Selected:
            painter.fillRect(option.rect, self.bgcolor)

        self.parent().setIndexWidget(index, index.data(Qt.DisplayRole))

    # ==================================================
    def sizeHint(self, option, index):
        item = index.data(Qt.DisplayRole)

        if item:
            return QSize(item.width(), item.height())
        else:
            return super().sizeHint(option, index)



import os
import pyvista as pv
import matplotlib.pyplot as plt
import seaborn as sns
from pyvistaqt import QtInteractor
from qtpy.QtWidgets import QGridLayout, QDialog, QLabel, QFileDialog
from qtpy.QtCore import QSize, Qt
from qtpy.uic import loadUi
from qtdraw import __version__, __author__, __date__
import warnings

warnings.filterwarnings("ignore")


# ==================================================
class QtDrawBase(QDialog):
    """
    base widget.
    """

    # ==================================================
    def __init__(
        self,
        parent=None,
        panel=None,
        theme="document",
        smooth_line=True,
        smooth_point=True,
        smooth_polygon=True,
        smooth_shading=True,
        sampling=10,
    ):
        """
        initialize the class.

        Args:
            parent (QObject, optional): parent object.
            panel (str, optional): UI file name for right panel.
            theme (str, optional): theme of pyvista, "document/paraview/default/dark".
            smooth_line (bool, optional): smoothing for line ?
            smooth_point (bool, optional): smoothing for point ?
            smooth_polygon (bool, optional): smoothing for polygon ?
            smooth_shading (bool, optional): smoothing for shading ?
            sampling (int, optional): the number of sampling for smoothing.
        """
        super().__init__(parent)
        self.shading = smooth_shading

        # set qt interactor.
        self._set_theme(theme)
        self._layer = QtInteractor(
            self,
            line_smoothing=smooth_line,
            point_smoothing=smooth_point,
            polygon_smoothing=smooth_polygon,
            multi_samples=sampling,
        )
        self._layer.setMinimumSize(QSize(200, 200))

        # set status box in the left bottom.
        self.status_box = QLabel(self)
        self.status_box.setContentsMargins(15, 5, 0, 5)

        # set copyright box in the right bottom.
        self.copyright_box = QLabel(self)
        self.copyright_box.setContentsMargins(0, 5, 15, 5)
        self.copyright_box.setAlignment(Qt.AlignRight)

        # layout.
        self.layout = QGridLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)

        self.layout.addWidget(self._layer.interactor, 0, 0, 1, 2)
        self.layout.addWidget(self.status_box, 1, 0, 1, 1)

        self.setLayout(self.layout)

        self._load_panel(panel)

        # set copyright.
        self.copyright_str = f"Versoin {__version__}, Copyright (C) {__date__} by {__author__}"
        self.set_copyright(self.copyright_str)

    # ==================================================
    def _set_theme(self, theme):
        """
        set pyvista theme.

        Args:
            theme (str): base theme.
        """
        pv.set_plot_theme(theme)
        pv.global_theme.transparent_background = True
        pv.global_theme.axes.show = False
        pv.global_theme.axes.box = False
        pv.global_theme.cmap = "bwr"
        pv.global_theme.show_scalar_bar = False

    # ==================================================
    def _set_anti_aliasing(self, mode=False):
        """
        set anti aliasing mode.

        Args:
            mode (bool, optional): anti aliaing mode.
        """
        if mode:
            self._layer.enable_anti_aliasing()
        else:
            self._layer.disable_anti_aliasing()

    # ==================================================
    def _set_parallel_projection(self, mode=True):
        """
        set parallel projection mode.

        Args:
            mode (bool, optional): parallel projection mode.
        """
        if mode:
            self._layer.enable_parallel_projection()
        else:
            self._layer.disable_parallel_projection()

    # ==================================================
    def _load_panel(self, fname):
        """
        load UI file for right panel.

        Args:
            fname (str): file name of UI.

        Notes:
            - if fname is None, no panel is used.
            - if fname is "", default panel is used.
        """
        if fname == "":
            fname = os.path.dirname(__file__) + "/core/default_panel.ui"
            fname = fname.replace(os.sep, "/")

        if fname is None:
            self.layout.addWidget(self.copyright_box, 1, 1, 1, 1)
        else:
            loadUi(fname, self)
            self.layout.addWidget(self.panel, 0, 2, 1, 1)
            self.layout.addWidget(self.copyright_box, 1, 1, 1, 2)

    # ==================================================
    def set_window_size(self, width=1174, height=768):
        """
        set window size.

        Args:
            width (int, optional): width.
            height (int, optional): height.
        """
        self.resize(width, height)

    # ==================================================
    def set_window_title(self, title):
        """
        set window title.

        Args:
            title (str): window title.
        """
        self.setWindowTitle(title)

    # ==================================================
    def set_copyright(self, text):
        """
        set copyright text.

        Args:
            text (str): copyright text in the right-bottom panel.
        """
        self.copyright_box.setText(text)

    # ==================================================
    def set_status(self, text):
        """
        set status text.

        Args:
            text (str): statu text in the left-bottom panel.
        """
        self.status_box.setText(text)

    # ==================================================
    def _screenshot(self, file_type=".png"):
        """
        screenshot dialog.

        Args:
            file_type (str, optional): file type, ".png/.bmp/.tif/.tiff/.svg/.eps/.ps/.pdf".
        """
        ifile = "*.png *.bmp *.tif *.tiff"  # jpeg, jpg do not work at present.
        gfile = "*.svg *.eps *.ps *.pdf"

        default_file = os.getcwd() + "/screenshot" + file_type
        default_file = default_file.replace(os.sep, "/")

        filename, _ = QFileDialog.getSaveFileName(
            self,
            "Save Screenshot",
            default_file,
            "Image files (" + ifile + ")",
            "Graphic files (" + gfile + ")",
        )

        if filename:
            _, ext = os.path.splitext(filename)
            if ext == "":
                ext = file_type
                filename += ext
            if ext in ifile.replace("*", "").split(" "):
                self._layer.screenshot(filename, transparent_background=True)
                self.set_status(f"Saved: {filename}")
            elif ext in gfile.replace("*", "").split(" "):
                self._layer.save_graphic(filename, "")
                self.set_status(f"Saved: {filename}")
            else:
                self.set_status(f"{ext} is unsupported.")

    # ==================================================
    def add_chart(self, chart, *charts):
        """
        add 2d chart.

        Args:
            chart (Chart): pyvsita chart.
        """
        self._layer.add_chart(chart, *charts)

    # ==================================================
    def chart(self):
        """
        chart.

        Returns:
            Chart2D: chart.
        """
        chart = pv.Chart2D()
        self._layer.add_chart(chart)

        return chart

    # ==================================================
    def chart_mpl(self, font_scale=1.0, width=1.0):
        """
        matplotlib chart.

        Returns:
            Figure: matplotlib figure.
        """
        sns.set(
            "notebook",
            "whitegrid",
            "dark",
            font_scale=1.2 * font_scale,
            rc={
                "lines.linewidth": width,
                "grid.linestyle": "--",
                "grid.linewidth": 0.6,
                "axes.linewidth": 0.6,
                "axes.xmargin": 0.0,
                "axes.ymargin": 0.0,
                "xtick.major.pad": 0.0,
                "xtick.minor.pad": 0.0,
                "ytick.major.pad": 0.0,
                "ytick.minor.pad": 0.0,
            },
        )

        figure = plt.figure(tight_layout=False)
        chart = pv.ChartMPL(figure)
        chart.background_color = "w"
        self._layer.add_chart(chart)

        return figure

    # ==================================================
    def _screen_off(self):
        """
        turn off rendering screen.
        """
        self._layer.ren_win.SetOffScreenRendering(1)

    # ==================================================
    def _screen_on(self):
        """
        turn on rendering screen.
        """
        self._layer.ren_win.SetOffScreenRendering(0)

    # ==================================================
    @property
    def layer(self):
        """
        3d plotting layer.

        Returns:
            QtInteractor: qt interactor to plot 3d object.
        """
        return self._layer

    # ==================================================
    def closeEvent(self, QCloseEvent):
        super().closeEvent(QCloseEvent)
        self._layer.close()
        self.close()
